# -*- coding: utf-8 -*-
"""Objects."""

from typing import TYPE_CHECKING, Callable, TypeVar, cast

try:
    import collections.abc as collections_abc
except ImportError:
    import collections as collections_abc  # type: ignore

from decorator import decorator
from six import string_types

from ._applications import Action
from ._bases import MISSING
from ._data import DataRelationship
from ._objects import (
    DATA_METHOD_TAG,
    Attribute,
    BaseReaction,
    DictObject,
    HistoryDescriptor,
    ListObject,
    MutableDictObject,
    MutableListObject,
    MutableSetObject,
    Object,
    ProxyDictObject,
    ProxyListObject,
    ProxySetObject,
    Relationship,
    SetObject,
)
from ._reactions import reaction
from ._structures import (
    KeyRelationship,
    UniqueDescriptor,
    make_auxiliary_cls,
    unique_descriptor,
)
from .utils.caller_module import get_caller_module
from .utils.factoring import import_factory
from .utils.reraise_context import ReraiseContext
from .utils.type_checking import assert_is_instance

if TYPE_CHECKING:
    from typing import Any, Dict, Iterable, Optional, Tuple, Type, Union

    from .utils.factoring import LazyFactory
    from .utils.type_checking import LazyTypes

    ReactionType = Union[LazyFactory, BaseReaction]
    ReactionsType = Union[ReactionType, Iterable[ReactionType]]

__all__ = [
    "Object",
    "DictObject",
    "MutableDictObject",
    "ListObject",
    "MutableListObject",
    "SetObject",
    "MutableSetObject",
    "ProxyDictObject",
    "ProxyListObject",
    "ProxySetObject",
    "Relationship",
    "Attribute",
    "KeyRelationship",
    "UniqueDescriptor",
    "Action",
    "data_method",
    "data_relationship",
    "unique_descriptor",
    "history_descriptor",
    "attribute",
    "constant_attribute",
    "protected_attribute_pair",
    "dict_attribute",
    "protected_dict_attribute_pair",
    "list_attribute",
    "protected_list_attribute_pair",
    "set_attribute",
    "protected_set_attribute_pair",
    "dict_cls",
    "list_cls",
    "set_cls",
]


T = TypeVar("T")  # Any type.
KT = TypeVar("KT")  # Any key type.
VT = TypeVar("VT")  # Any value type.


if TYPE_CHECKING:
    NT = Union[Type[None], None]
    MutableDictAttribute = Attribute[MutableDictObject[KT, VT]]
    MutableListAttribute = Attribute[MutableListObject[T]]
    MutableSetAttribute = Attribute[MutableSetObject[T]]
    DictAttribute = Attribute[DictObject[KT, VT]]
    ListAttribute = Attribute[ListObject[T]]
    SetAttribute = Attribute[SetObject[T]]
    ProxyDictAttribute = Attribute[ProxyDictObject[KT, VT]]
    ProxyListAttribute = Attribute[ProxyListObject[T]]
    ProxySetAttribute = Attribute[ProxySetObject[T]]


# noinspection PyAbstractClass
def data_method(func):
    # type: (Callable) -> Callable
    """
    Decorate object methods by tagging them as data methods.
    The generated data class will have the decorated methods in them.

    .. code:: python

        >>> from objetto.applications import Application
        >>> from objetto.objects import Object, attribute, data_method

        >>> class MyObject(Object):
        ...     value = attribute(int, default=0)
        ...
        ...     @data_method
        ...     def get_double(self):
        ...         return self.value * 2
        ...
        >>> app = Application()
        >>> my_obj = MyObject(app)
        >>> my_obj.value = 42
        >>> my_obj.get_double()
        84
        >>> my_obj.data.get_double()
        84

    :param func: The method function.
    :type func: function

    :return: Decorated method function.
    :rtype: function
    """

    @decorator
    def data_method_(func_, *args, **kwargs):
        """Data method decorator."""
        return func_(*args, **kwargs)

    decorated = data_method_(func)
    setattr(decorated, DATA_METHOD_TAG, True)
    return decorated


def data_relationship(
    types=(),  # type: LazyTypes
    subtypes=False,  # type: bool
    checked=None,  # type: Optional[bool]
    module=None,  # type: Optional[str]
    factory=None,  # type: LazyFactory
    serialized=True,  # type: bool
    serializer=None,  # type: LazyFactory
    deserializer=None,  # type: LazyFactory
    represented=True,  # type: bool
    compared=True,  # type: bool
):
    # type: (...) -> DataRelationship
    """
    Make custom relationship between a data structure and its values.

    :param types: Types.
    :type types: str or type or None or tuple[str or type or None]

    :param subtypes: Whether to accept subtypes.
    :type subtypes: bool

    :param checked: Whether to perform runtime type check.
    :type checked: bool or None

    :param module: Module path for lazy types/factories.
    :type module: str or None

    :param factory: Value factory.
    :type factory: str or collections.abc.Callable or None

    :param serialized: Whether should be serialized.
    :type serialized: bool

    :param serializer: Custom serializer.
    :type serializer: str or collections.abc.Callable or None

    :param deserializer: Custom deserializer.
    :type deserializer: str or collections.abc.Callable or None

    :param represented: Whether should be represented.
    :type represented: bool

    :param compared: Whether the value should be leverage when comparing.
    :type compared: bool

    :return: Custom data relationship.
    :rtype: objetto.data.DataRelationship

    :raises TypeError: Invalid parameter type.
    :raises ValueError: Invalid parameter value.
    """
    return DataRelationship(
        types=types,
        subtypes=subtypes,
        checked=checked,
        module=module,
        factory=factory,
        serialized=serialized,
        serializer=serializer,
        deserializer=deserializer,
        represented=represented,
        compared=compared,
    )


# noinspection PyAbstractClass
def history_descriptor(size=None):
    # type: (Optional[int]) -> HistoryDescriptor
    """
    Descriptor to be used when declaring an :class:`objetto.objects.Object` class.

    When used, every instance of the object class will hold a history that will keep
    track of its changes (and the changes of its children that define a history
    relationship), allowing for easy undo/redo operations.
    If accessed through an instance, the descriptor will return the history object.

    .. code:: python

        >>> from objetto import Application, Object, attribute, history_descriptor

        >>> class Person(Object):
        ...     history = history_descriptor()
        ...     name = attribute(str)
        ...
        >>> app = Application()
        >>> person = Person(app, name="Albert")
        >>> person.name = "Einstein"
        >>> person.history.undo()
        >>> person.name
        'Albert'

    :param size: How many changes to remember.
    :type size: int or None

    :return: History descriptor.
    :rtype: objetto.history.HistoryDescriptor
    """
    return HistoryDescriptor(size=size)


def attribute(
    types=(),  # type: Union[Type[T], NT, str, Iterable[Union[Type[T], NT, str]]]
    subtypes=False,  # type: bool
    checked=None,  # type: Optional[bool]
    module=None,  # type: Optional[str]
    factory=None,  # type: LazyFactory
    serialized=None,  # type: Optional[bool]
    serializer=None,  # type: LazyFactory
    deserializer=None,  # type: LazyFactory
    represented=True,  # type: bool
    child=True,  # type: bool
    history=None,  # type: Optional[bool]
    data=None,  # type: Optional[bool]
    custom_data_relationship=None,  # type: Optional[DataRelationship]
    default=MISSING,  # type: Any
    default_factory=None,  # type: LazyFactory
    required=True,  # type: bool
    changeable=None,  # type: Optional[bool]
    deletable=None,  # type: Optional[bool]
    finalized=False,  # type: bool
    abstracted=False,  # type: bool
    delegated=False,  # type: bool
    dependencies=None,  # type: Optional[Union[Iterable[Attribute], Attribute]]
    deserialize_to=None,  # type: Optional[Attribute]
):
    # type: (...) -> Attribute[T]
    """
    Make attribute.

    :param types: Types.
    :type types: str or type or None or tuple[str or type or None]

    :param subtypes: Whether to accept subtypes.
    :type subtypes: bool

    :param checked: Whether to perform runtime type check.
    :type checked: bool

    :param module: Module path for lazy types/factories.
    :type module: str or None

    :param factory: Value factory.
    :type factory: str or collections.abc.Callable or None

    :param serialized: Whether should be serialized.
    :type serialized: bool

    :param serializer: Custom serializer.
    :type serializer: str or collections.abc.Callable or None

    :param deserializer: Custom deserializer.
    :type deserializer: str or collections.abc.Callable or None

    :param represented: Whether should be represented.
    :type represented: bool

    :param child: Whether object values should be adopted as children.
    :type child: bool

    :param history: Whether to propagate the history to the child object value.
    :type history: bool

    :param data: Whether to generate data for the value.
    :type data: bool

    :param custom_data_relationship: Custom data relationship.
    :type custom_data_relationship: objetto.data.DataRelationship or None

    :param default: Default value.

    :param default_factory: Default value factory.
    :type default_factory: str or collections.abc.Callable or None

    :param required: Whether attribute is required to have a value or not.
    :type required: bool

    :param changeable: Whether attribute value can be changed.
    :type changeable: bool

    :param deletable: Whether attribute value can be deleted.
    :type deletable: bool

    :param finalized: If True, attribute can't be overridden by subclasses.
    :type finalized: bool

    :param abstracted: If True, attribute needs to be overridden by subclasses.
    :type abstracted: bool

    :param delegated: Whether attribute allows for delegates to be defined.
    :type delegated: bool

    :param dependencies: Attributes needed by the getter delegate.
    :type dependencies: collections.abc.Iterable[objetto.objects.Attribute] or \
objetto.objects.Attribute or None

    :param deserialize_to: Non-serialized attribute to deserialize this into.
    :type deserialize_to: objetto.objects.Attribute or None

    :return: Attribute.
    :rtype: objetto.objects.Attribute

    :raises TypeError: Invalid parameter type.
    :raises ValueError: Invalid parameter value.
    :raises ValueError: Can't declare same dependency more than once.
    :raises ValueError: Provided 'changeable' but 'delegated' is True.
    :raises ValueError: Provided 'deletable' but 'delegated' is True.
    :raises ValueError: Provided 'dependencies' but 'delegated' is False.
    :raises ValueError: Provided 'deserialize_to' but 'serialized' is False.
    :raises ValueError: Can't provide a serialized attribute to 'deserialize_to'.
    """

    # Get module from caller if not provided.
    module = get_caller_module() if module is None else module

    # Relationship.
    with ReraiseContext((TypeError, ValueError), "defining 'attribute'"):
        if custom_data_relationship is not None:
            assert_is_instance(
                custom_data_relationship, DataRelationship, subtypes=False
            )
        relationship = Relationship(
            types=types,
            subtypes=subtypes,
            checked=checked,
            module=module,
            factory=factory,
            serialized=serialized,
            serializer=serializer,
            deserializer=deserializer,
            represented=represented,
            child=child,
            history=history,
            data=data,
            data_relationship=custom_data_relationship,
        )

    # Make attribute.
    with ReraiseContext((TypeError, ValueError), "defining 'attribute'"):
        attribute_ = Attribute(
            relationship=relationship,
            default=default,
            default_factory=default_factory,
            module=module,
            required=required,
            changeable=changeable,
            deletable=deletable,
            finalized=finalized,
            abstracted=abstracted,
            delegated=delegated,
            dependencies=dependencies,
            deserialize_to=deserialize_to,
        )  # type: Attribute[T]

    return attribute_


def constant_attribute(
    value,  # type: T
    serialized=False,  # type: bool
    serializer=None,  # type: LazyFactory
    deserializer=None,  # type: LazyFactory
    represented=False,  # type: bool
    data=True,  # type: bool
    finalized=False,  # type: bool
    abstracted=False,  # type: bool
):
    # type: (...) -> Attribute[T]
    """
    Make constant attribute.

    :param value: Constant value.

    :param serialized: Whether should be serialized.
    :type serialized: bool

    :param serializer: Custom serializer.
    :type serializer: str or collections.abc.Callable or None

    :param deserializer: Custom deserializer.
    :type deserializer: str or collections.abc.Callable or None

    :param represented: Whether should be represented.
    :type represented: bool

    :param data: Whether to generate data for the value.
    :type data: bool

    :param finalized: If True, attribute can't be overridden by subclasses.
    :type finalized: bool

    :param abstracted: If True, attribute needs to be overridden by subclasses.
    :type abstracted: bool

    :return: Constant attribute.
    :rtype: objetto.objects.Attribute

    :raises TypeError: Invalid parameter type.
    :raises ValueError: Invalid parameter value.
    """

    # Relationship.
    with ReraiseContext((TypeError, ValueError), "defining 'constant_attribute'"):
        relationship = Relationship(
            types=type(value),
            subtypes=False,
            checked=False,
            serialized=serialized,
            serializer=serializer,
            deserializer=deserializer,
            represented=represented,
            child=True,
            history=False,
            data=data,
            data_relationship=None,
        )

    # Make attribute.
    with ReraiseContext((TypeError, ValueError), "defining 'constant_attribute'"):
        attribute_ = Attribute(
            relationship=relationship,
            default=value,
            default_factory=None,
            required=False,
            changeable=False,
            deletable=False,
            finalized=finalized,
            abstracted=abstracted,
        )  # type: Attribute[T]

    return attribute_


def protected_attribute_pair(
    types=(),  # type: Union[Type[T], NT, str, Iterable[Union[Type[T], NT, str]]]
    subtypes=False,  # type: bool
    checked=None,  # type: Optional[bool]
    module=None,  # type: Optional[str]
    factory=None,  # type: LazyFactory
    serialized=None,  # type: Optional[bool]
    serializer=None,  # type: LazyFactory
    deserializer=None,  # type: LazyFactory
    represented=True,  # type: bool
    child=True,  # type: bool
    history=None,  # type: Optional[bool]
    data=None,  # type: Optional[bool]
    custom_data_relationship=None,  # type: Optional[DataRelationship]
    default=MISSING,  # type: Any
    default_factory=None,  # type: LazyFactory
    changeable=None,  # type: Optional[bool]
    deletable=None,  # type: Optional[bool]
    finalized=False,  # type: bool
    abstracted=False,  # type: bool
):
    # type: (...) -> Tuple[Attribute[T], Attribute[T]]
    """
    Make protected-public attribute pair.

    :param types: Types.
    :type types: str or type or None or tuple[str or type or None]

    :param subtypes: Whether to accept subtypes.
    :type subtypes: bool

    :param checked: Whether to perform runtime type check.
    :type checked: bool

    :param module: Module path for lazy types/factories.
    :type module: str or None

    :param factory: Value factory.
    :type factory: str or collections.abc.Callable or None

    :param serialized: Whether should be serialized.
    :type serialized: bool

    :param serializer: Custom serializer.
    :type serializer: str or collections.abc.Callable or None

    :param deserializer: Custom deserializer.
    :type deserializer: str or collections.abc.Callable or None

    :param represented: Whether should be represented.
    :type represented: bool

    :param child: Whether object values should be adopted as children.
    :type child: bool

    :param history: Whether to propagate the history to the child object value.
    :type history: bool

    :param data: Whether to generate data for the value.
    :type data: bool

    :param custom_data_relationship: Custom data relationship.
    :type custom_data_relationship: objetto.data.DataRelationship or None

    :param default: Default value.

    :param default_factory: Default value factory.
    :type default_factory: str or collections.abc.Callable or None

    :param changeable: Whether attribute value can be changed.
    :type changeable: bool

    :param deletable: Whether attribute value can be deleted.
    :type deletable: bool

    :param finalized: If True, attribute can't be overridden by subclasses.
    :type finalized: bool

    :param abstracted: If True, attribute needs to be overridden by subclasses.
    :type abstracted: bool

    :return: Protected-public attribute pair.
    :rtype: tuple[objetto.objects.Attribute, objetto.objects.Attribute]

    :raises TypeError: Invalid parameter type.
    :raises ValueError: Invalid parameter value.
    """

    # Get module from caller if not provided.
    module = get_caller_module() if module is None else module

    # Make protected attribute.
    protected_attribute = attribute(
        types=types,
        subtypes=subtypes,
        checked=checked,
        module=module,
        factory=factory,
        represented=False,
        child=False,
        default=default,
        default_factory=default_factory,
        required=False,
        changeable=changeable,
        deletable=deletable,
        finalized=finalized,
        abstracted=abstracted,
        delegated=False,
    )

    # Make public attribute.
    public_attribute = attribute(
        types=types,
        subtypes=subtypes,
        checked=False,
        module=module,
        serialized=serialized,
        serializer=serializer,
        deserializer=deserializer,
        represented=represented,
        child=child,
        history=history,
        data=data,
        custom_data_relationship=custom_data_relationship,
        required=False,
        finalized=finalized,
        abstracted=abstracted,
        delegated=True,
        dependencies=(protected_attribute,),
        deserialize_to=protected_attribute if serialized else None,
    )

    def getter(iobj):
        """Getter delegate."""
        protected_name = iobj.__.cls._attribute_names[protected_attribute]
        return iobj[protected_name]

    public_attribute.getter(getter)

    return protected_attribute, public_attribute


def dict_attribute(
    types=(),  # type: Union[Type[VT], NT, str, Iterable[Union[Type[VT], NT, str]]]
    subtypes=False,  # type: bool
    checked=None,  # type: Optional[bool]
    module=None,  # type: Optional[str]
    factory=None,  # type: LazyFactory
    serialized=None,  # type: Optional[bool]
    serializer=None,  # type: LazyFactory
    deserializer=None,  # type: LazyFactory
    represented=True,  # type: bool
    key_types=(),  # type: Union[Type[KT], NT, str, Iterable[Union[Type[KT], NT, str]]]
    key_subtypes=False,  # type: bool
    key_factory=None,  # type: LazyFactory
    child=True,  # type: bool
    history=None,  # type: Optional[bool]
    data=None,  # type: Optional[bool]
    custom_data_relationship=None,  # type: Optional[DataRelationship]
    default=MISSING,  # type: Any
    default_factory=None,  # type: LazyFactory
    required=False,  # type: bool
    changeable=True,  # type: bool
    deletable=False,  # type: bool
    finalized=False,  # type: bool
    abstracted=False,  # type: bool
    qual_name=None,  # type: Optional[str]
    unique=False,  # type: bool
    reactions=None,  # type: ReactionsType
    mutable=True,  # type: bool
):
    # type: (...) -> Union[MutableDictAttribute[KT, VT], DictAttribute[KT, VT]]
    """
    Make dictionary attribute.

    :param types: Types.
    :type types: str or type or None or tuple[str or type or None]

    :param subtypes: Whether to accept subtypes.
    :type subtypes: bool

    :param checked: Whether to perform runtime type check.
    :type checked: bool

    :param module: Module path for lazy types/factories.
    :type module: str or None

    :param factory: Value factory.
    :type factory: str or collections.abc.Callable or None

    :param serialized: Whether should be serialized.
    :type serialized: bool

    :param serializer: Custom serializer.
    :type serializer: str or collections.abc.Callable or None

    :param deserializer: Custom deserializer.
    :type deserializer: str or collections.abc.Callable or None

    :param represented: Whether should be represented.
    :type represented: bool

    :param child: Whether object values should be adopted as children.
    :type child: bool

    :param history: Whether to propagate the history to the child object value.
    :type history: bool

    :param data: Whether to generate data for the value.
    :type data: bool

    :param custom_data_relationship: Custom data relationship.
    :type custom_data_relationship: objetto.data.DataRelationship or None

    :param key_types: Key types.
    :type key_types: str or type or None or tuple[str or type or None]

    :param key_subtypes: Whether to accept subtypes for the keys.
    :type key_subtypes: bool

    :param key_factory: Key factory.
    :type key_factory: str or collections.abc.Callable or None

    :param default: Default value.

    :param default_factory: Default value factory.
    :type default_factory: str or collections.abc.Callable or None

    :param required: Whether attribute is required to have a value or not.
    :type required: bool

    :param changeable: Whether attribute value can be changed.
    :type changeable: bool

    :param deletable: Whether attribute value can be deleted.
    :type deletable: bool

    :param finalized: If True, attribute can't be overridden by subclasses.
    :type finalized: bool

    :param abstracted: If True, attribute needs to be overridden by subclasses.
    :type abstracted: bool

    :param qual_name: Optional type qualified name for the generated class.
    :type qual_name: str or None

    :param unique: Whether generated class should have a unique descriptor.
    :type unique: bool

    :param reactions: Reaction functions ordered by priority.
    :type reactions: str or collections.abc.Callable or None or \
collections.abc.Iterable[str or collections.abc.Callable]

    :param mutable: Whether generated class should have public mutable interface.
    :type mutable: bool

    :return: Dictionary attribute.
    :rtype: objetto.objects.Attribute[objetto.objects.MutableDictObject or \
objetto.objects.DictObject]

    :raises TypeError: Invalid parameter type.
    :raises ValueError: Invalid parameter value.
    """

    # Get module from caller if not provided.
    module = get_caller_module() if module is None else module

    # Make dictionary class.
    with ReraiseContext((TypeError, ValueError), "defining 'dict_attribute'"):
        dict_type = dict_cls(
            types=types,
            subtypes=subtypes,
            checked=checked,
            module=module,
            factory=factory,
            serialized=serialized,
            serializer=serializer,
            deserializer=deserializer,
            represented=represented,
            key_types=key_types,
            key_subtypes=key_subtypes,
            key_factory=key_factory,
            child=child,
            history=history,
            data=data,
            custom_data_relationship=custom_data_relationship,
            qual_name=qual_name,
            unique=unique,
            reactions=reactions,
            mutable=mutable,
        )  # type: Union[Type[MutableDictObject[KT, VT]], Type[DictObject[KT, VT]]]

    # Factory for dict object relationship.
    def dict_factory(initial=(), app=None, **_):
        """Factory for the whole dict object."""
        if type(initial) is dict_type and initial.app is app:
            return initial
        else:
            return dict_type(app, initial)

    # Get default value/factory.
    if changeable and not required and default is MISSING and default_factory is None:
        default = ()

    # Relationship.
    with ReraiseContext((TypeError, ValueError), "defining 'attribute'"):
        relationship = Relationship(
            types=dict_type,
            subtypes=False,
            checked=False,
            module=module,
            factory=dict_factory,
            serialized=serialized,
            serializer=None,
            deserializer=None,
            represented=represented,
            child=child,
            history=history,
            data=data,
            data_relationship=None,
        )

    # Make attribute.
    with ReraiseContext((TypeError, ValueError), "defining 'attribute'"):
        attribute_ = Attribute(
            relationship=relationship,
            default=default,
            default_factory=default_factory,
            module=module,
            required=required,
            changeable=changeable,
            deletable=deletable,
            finalized=finalized,
            abstracted=abstracted,
            delegated=False,
            dependencies=None,
            deserialize_to=None,
        )  # type: Attribute[MutableDictObject[KT, VT]]

    return attribute_


def protected_dict_attribute_pair(
    types=(),  # type: Union[Type[VT], NT, str, Iterable[Union[Type[VT], NT, str]]]
    subtypes=False,  # type: bool
    checked=None,  # type: Optional[bool]
    module=None,  # type: Optional[str]
    factory=None,  # type: LazyFactory
    serialized=None,  # type: Optional[bool]
    serializer=None,  # type: LazyFactory
    deserializer=None,  # type: LazyFactory
    represented=True,  # type: bool
    key_types=(),  # type: Union[Type[KT], NT, str, Iterable[Union[Type[KT], NT, str]]]
    key_subtypes=False,  # type: bool
    key_factory=None,  # type: LazyFactory
    child=True,  # type: bool
    history=None,  # type: Optional[bool]
    data=None,  # type: Optional[bool]
    custom_data_relationship=None,  # type: Optional[DataRelationship]
    default=MISSING,  # type: Any
    default_factory=None,  # type: LazyFactory
    finalized=False,  # type: bool
    abstracted=False,  # type: bool
    qual_name=None,  # type: Optional[str]
    unique=False,  # type: bool
    reactions=None,  # type: ReactionsType
):
    # type: (...) -> Tuple[ProxyDictAttribute[KT, VT], DictAttribute[KT, VT]]
    """
    Make protected-public dictionary attribute pair.

    :param types: Types.
    :type types: str or type or None or tuple[str or type or None]

    :param subtypes: Whether to accept subtypes.
    :type subtypes: bool

    :param checked: Whether to perform runtime type check.
    :type checked: bool

    :param module: Module path for lazy types/factories.
    :type module: str or None

    :param factory: Value factory.
    :type factory: str or collections.abc.Callable or None

    :param serialized: Whether should be serialized.
    :type serialized: bool

    :param serializer: Custom serializer.
    :type serializer: str or collections.abc.Callable or None

    :param deserializer: Custom deserializer.
    :type deserializer: str or collections.abc.Callable or None

    :param represented: Whether should be represented.
    :type represented: bool

    :param child: Whether object values should be adopted as children.
    :type child: bool

    :param history: Whether to propagate the history to the child object value.
    :type history: bool

    :param data: Whether to generate data for the value.
    :type data: bool

    :param custom_data_relationship: Custom data relationship.
    :type custom_data_relationship: objetto.data.DataRelationship or None

    :param key_types: Key types.
    :type key_types: str or type or None or tuple[str or type or None]

    :param key_subtypes: Whether to accept subtypes for the keys.
    :type key_subtypes: bool

    :param key_factory: Key factory.
    :type key_factory: str or collections.abc.Callable or None

    :param default: Default value.

    :param default_factory: Default value factory.
    :type default_factory: str or collections.abc.Callable or None

    :param finalized: If True, attribute can't be overridden by subclasses.
    :type finalized: bool

    :param abstracted: If True, attribute needs to be overridden by subclasses.
    :type abstracted: bool

    :param qual_name: Optional type qualified name for the generated class.
    :type qual_name: str or None

    :param unique: Whether generated class should have a unique descriptor.
    :type unique: bool

    :param reactions: Reaction functions ordered by priority.
    :type reactions: str or collections.abc.Callable or None or \
collections.abc.Iterable[str or collections.abc.Callable]

    :return: Protected-public dictionary attribute pair.
    :rtype: tuple[objetto.objects.Attribute[objetto.objects.ProxyDictObject], \
objetto.objects.Attribute[objetto.objects.DictObject]]

    :raises TypeError: Invalid parameter type.
    :raises ValueError: Invalid parameter value.
    """

    # Get module from caller if not provided.
    module = get_caller_module() if module is None else module

    # Make public dictionary attribute.
    if default is MISSING and default_factory is None:
        default = ()

    public_attribute = cast(
        "DictAttribute[KT, VT]",
        dict_attribute(
            types=types,
            subtypes=subtypes,
            checked=checked,
            module=module,
            factory=factory,
            serialized=serialized,
            serializer=serializer,
            deserializer=deserializer,
            represented=represented,
            key_types=key_types,
            key_subtypes=key_subtypes,
            key_factory=key_factory,
            child=child,
            history=history,
            data=data,
            custom_data_relationship=custom_data_relationship,
            default=default,
            default_factory=default_factory,
            required=False,
            changeable=False,
            deletable=False,
            finalized=finalized,
            abstracted=abstracted,
            qual_name=qual_name,
            unique=unique,
            reactions=reactions,
            mutable=False,
        ),
    )

    # Make protected attribute.
    protected_attribute = cast(
        "ProxyDictAttribute[KT, VT]",
        attribute(
            types=ProxyDictObject,
            subtypes=False,
            checked=False,
            module=module,
            represented=False,
            child=False,
            required=False,
            finalized=finalized,
            abstracted=abstracted,
            delegated=True,
            dependencies=(public_attribute,),
        ),
    )

    def getter(iobj):
        """Getter delegate."""
        public_name = iobj.__.cls._attribute_names[public_attribute]
        return ProxyDictObject(iobj[public_name])

    protected_attribute.getter(getter)

    return protected_attribute, public_attribute


def list_attribute(
    types=(),  # type: Union[Type[T], NT, str, Iterable[Union[Type[T], NT, str]]]
    subtypes=False,  # type: bool
    checked=None,  # type: Optional[bool]
    module=None,  # type: Optional[str]
    factory=None,  # type: LazyFactory
    serialized=None,  # type: Optional[bool]
    serializer=None,  # type: LazyFactory
    deserializer=None,  # type: LazyFactory
    represented=True,  # type: bool
    child=True,  # type: bool
    history=None,  # type: Optional[bool]
    data=None,  # type: Optional[bool]
    custom_data_relationship=None,  # type: Optional[DataRelationship]
    default=MISSING,  # type: Any
    default_factory=None,  # type: LazyFactory
    required=False,  # type: bool
    changeable=True,  # type: bool
    deletable=False,  # type: bool
    finalized=False,  # type: bool
    abstracted=False,  # type: bool
    qual_name=None,  # type: Optional[str]
    unique=False,  # type: bool
    reactions=None,  # type: ReactionsType
    mutable=True,  # type: bool
):
    # type: (...) -> Union[MutableListAttribute[T], ListAttribute[T]]
    """
    Make list attribute.

    :param types: Types.
    :type types: str or type or None or tuple[str or type or None]

    :param subtypes: Whether to accept subtypes.
    :type subtypes: bool

    :param checked: Whether to perform runtime type check.
    :type checked: bool

    :param module: Module path for lazy types/factories.
    :type module: str or None

    :param factory: Value factory.
    :type factory: str or collections.abc.Callable or None

    :param serialized: Whether should be serialized.
    :type serialized: bool

    :param serializer: Custom serializer.
    :type serializer: str or collections.abc.Callable or None

    :param deserializer: Custom deserializer.
    :type deserializer: str or collections.abc.Callable or None

    :param represented: Whether should be represented.
    :type represented: bool

    :param child: Whether object values should be adopted as children.
    :type child: bool

    :param history: Whether to propagate the history to the child object value.
    :type history: bool

    :param data: Whether to generate data for the value.
    :type data: bool

    :param custom_data_relationship: Custom data relationship.
    :type custom_data_relationship: objetto.data.DataRelationship or None

    :param default: Default value.

    :param default_factory: Default value factory.
    :type default_factory: str or collections.abc.Callable or None

    :param required: Whether attribute is required to have a value or not.
    :type required: bool

    :param changeable: Whether attribute value can be changed.
    :type changeable: bool

    :param deletable: Whether attribute value can be deleted.
    :type deletable: bool

    :param finalized: If True, attribute can't be overridden by subclasses.
    :type finalized: bool

    :param abstracted: If True, attribute needs to be overridden by subclasses.
    :type abstracted: bool

    :param qual_name: Optional type qualified name for the generated class.
    :type qual_name: str or None

    :param unique: Whether generated class should have a unique descriptor.
    :type unique: bool

    :param reactions: Reaction functions ordered by priority.
    :type reactions: str or collections.abc.Callable or None or \
collections.abc.Iterable[str or collections.abc.Callable]

    :param mutable: Whether generated class should have public mutable interface.
    :type mutable: bool

    :return: List attribute.
    :rtype: objetto.objects.Attribute[objetto.objects.MutableListObject or \
objetto.objects.ListObject]

    :raises TypeError: Invalid parameter type.
    :raises ValueError: Invalid parameter value.
    """

    # Get module from caller if not provided.
    module = get_caller_module() if module is None else module

    # Make list class.
    with ReraiseContext((TypeError, ValueError), "defining 'list_attribute'"):
        list_type = list_cls(
            types=types,
            subtypes=subtypes,
            checked=checked,
            module=module,
            factory=factory,
            serialized=serialized,
            serializer=serializer,
            deserializer=deserializer,
            represented=represented,
            child=child,
            history=history,
            data=data,
            custom_data_relationship=custom_data_relationship,
            qual_name=qual_name,
            unique=unique,
            reactions=reactions,
            mutable=mutable,
        )  # type: Union[Type[MutableListObject[T]], Type[ListObject[T]]]

    # Factory for list object relationship.
    def list_factory(initial=(), app=None, **_):
        """Factory for the whole list object."""
        if type(initial) is list_type and initial.app is app:
            return initial
        else:
            return list_type(app, initial)

    # Get default value/factory.
    if changeable and not required and default is MISSING and default_factory is None:
        default = ()

    # Relationship.
    with ReraiseContext((TypeError, ValueError), "defining 'attribute'"):
        relationship = Relationship(
            types=list_type,
            subtypes=False,
            checked=False,
            module=module,
            factory=list_factory,
            serialized=serialized,
            serializer=None,
            deserializer=None,
            represented=represented,
            child=child,
            history=history,
            data=data,
            data_relationship=None,
        )

    # Make attribute.
    with ReraiseContext((TypeError, ValueError), "defining 'attribute'"):
        attribute_ = Attribute(
            relationship=relationship,
            default=default,
            default_factory=default_factory,
            module=module,
            required=required,
            changeable=changeable,
            deletable=deletable,
            finalized=finalized,
            abstracted=abstracted,
            delegated=False,
            dependencies=None,
            deserialize_to=None,
        )  # type: Attribute[MutableListObject[T]]

    return attribute_


def protected_list_attribute_pair(
    types=(),  # type: Union[Type[VT], NT, str, Iterable[Union[Type[VT], NT, str]]]
    subtypes=False,  # type: bool
    checked=None,  # type: Optional[bool]
    module=None,  # type: Optional[str]
    factory=None,  # type: LazyFactory
    serialized=None,  # type: Optional[bool]
    serializer=None,  # type: LazyFactory
    deserializer=None,  # type: LazyFactory
    represented=True,  # type: bool
    child=True,  # type: bool
    history=None,  # type: Optional[bool]
    data=None,  # type: Optional[bool]
    custom_data_relationship=None,  # type: Optional[DataRelationship]
    default=MISSING,  # type: Any
    default_factory=None,  # type: LazyFactory
    finalized=False,  # type: bool
    abstracted=False,  # type: bool
    qual_name=None,  # type: Optional[str]
    unique=False,  # type: bool
    reactions=None,  # type: ReactionsType
):
    # type: (...) -> Tuple[ProxyListAttribute[T], ListAttribute[T]]
    """
    Make protected-public list attribute pair.

    :param types: Types.
    :type types: str or type or None or tuple[str or type or None]

    :param subtypes: Whether to accept subtypes.
    :type subtypes: bool

    :param checked: Whether to perform runtime type check.
    :type checked: bool

    :param module: Module path for lazy types/factories.
    :type module: str or None

    :param factory: Value factory.
    :type factory: str or collections.abc.Callable or None

    :param serialized: Whether should be serialized.
    :type serialized: bool

    :param serializer: Custom serializer.
    :type serializer: str or collections.abc.Callable or None

    :param deserializer: Custom deserializer.
    :type deserializer: str or collections.abc.Callable or None

    :param represented: Whether should be represented.
    :type represented: bool

    :param child: Whether object values should be adopted as children.
    :type child: bool

    :param history: Whether to propagate the history to the child object value.
    :type history: bool

    :param data: Whether to generate data for the value.
    :type data: bool

    :param custom_data_relationship: Custom data relationship.
    :type custom_data_relationship: objetto.data.DataRelationship or None

    :param default: Default value.

    :param default_factory: Default value factory.
    :type default_factory: str or collections.abc.Callable or None

    :param finalized: If True, attribute can't be overridden by subclasses.
    :type finalized: bool

    :param abstracted: If True, attribute needs to be overridden by subclasses.
    :type abstracted: bool

    :param qual_name: Optional type qualified name for the generated class.
    :type qual_name: str or None

    :param unique: Whether generated class should have a unique descriptor.
    :type unique: bool

    :param reactions: Reaction functions ordered by priority.
    :type reactions: str or collections.abc.Callable or None or \
collections.abc.Iterable[str or collections.abc.Callable]

    :return: Protected-public list attribute pair.
    :rtype: tuple[objetto.objects.Attribute[objetto.objects.ProxyListObject], \
objetto.objects.Attribute[objetto.objects.ListObject]]

    :raises TypeError: Invalid parameter type.
    :raises ValueError: Invalid parameter value.
    """

    # Get module from caller if not provided.
    module = get_caller_module() if module is None else module

    # Make public list attribute.
    if default is MISSING and default_factory is None:
        default = ()

    public_attribute = cast(
        "ListAttribute[T]",
        list_attribute(
            types=types,
            subtypes=subtypes,
            checked=checked,
            module=module,
            factory=factory,
            serialized=serialized,
            serializer=serializer,
            deserializer=deserializer,
            represented=represented,
            child=child,
            history=history,
            data=data,
            custom_data_relationship=custom_data_relationship,
            default=default,
            default_factory=default_factory,
            required=False,
            changeable=False,
            deletable=False,
            finalized=finalized,
            abstracted=abstracted,
            qual_name=qual_name,
            unique=unique,
            reactions=reactions,
            mutable=False,
        ),
    )

    # Make protected attribute.
    protected_attribute = cast(
        "ProxyListAttribute[T]",
        attribute(
            types=ProxyListObject,
            subtypes=False,
            checked=False,
            module=module,
            represented=False,
            child=False,
            required=False,
            finalized=finalized,
            abstracted=abstracted,
            delegated=True,
            dependencies=(public_attribute,),
        ),
    )

    def getter(iobj):
        """Getter delegate."""
        public_name = iobj.__.cls._attribute_names[public_attribute]
        return ProxyListObject(iobj[public_name])

    protected_attribute.getter(getter)

    return protected_attribute, public_attribute


def set_attribute(
    types=(),  # type: Union[Type[T], NT, str, Iterable[Union[Type[T], NT, str]]]
    subtypes=False,  # type: bool
    checked=None,  # type: Optional[bool]
    module=None,  # type: Optional[str]
    factory=None,  # type: LazyFactory
    serialized=None,  # type: Optional[bool]
    serializer=None,  # type: LazyFactory
    deserializer=None,  # type: LazyFactory
    represented=True,  # type: bool
    child=True,  # type: bool
    history=None,  # type: Optional[bool]
    data=None,  # type: Optional[bool]
    custom_data_relationship=None,  # type: Optional[DataRelationship]
    default=MISSING,  # type: Any
    default_factory=None,  # type: LazyFactory
    required=False,  # type: bool
    changeable=True,  # type: bool
    deletable=False,  # type: bool
    finalized=False,  # type: bool
    abstracted=False,  # type: bool
    qual_name=None,  # type: Optional[str]
    unique=False,  # type: bool
    reactions=None,  # type: ReactionsType
    mutable=True,  # type: bool
):
    # type: (...) -> Union[MutableSetAttribute[T], SetAttribute[T]]
    """
    Make set attribute.

    :param types: Types.
    :type types: str or type or None or tuple[str or type or None]

    :param subtypes: Whether to accept subtypes.
    :type subtypes: bool

    :param checked: Whether to perform runtime type check.
    :type checked: bool

    :param module: Module path for lazy types/factories.
    :type module: str or None

    :param factory: Value factory.
    :type factory: str or collections.abc.Callable or None

    :param serialized: Whether should be serialized.
    :type serialized: bool

    :param serializer: Custom serializer.
    :type serializer: str or collections.abc.Callable or None

    :param deserializer: Custom deserializer.
    :type deserializer: str or collections.abc.Callable or None

    :param represented: Whether should be represented.
    :type represented: bool

    :param child: Whether object values should be adopted as children.
    :type child: bool

    :param history: Whether to propagate the history to the child object value.
    :type history: bool

    :param data: Whether to generate data for the value.
    :type data: bool

    :param custom_data_relationship: Custom data relationship.
    :type custom_data_relationship: objetto.data.DataRelationship or None

    :param default: Default value.

    :param default_factory: Default value factory.
    :type default_factory: str or collections.abc.Callable or None

    :param required: Whether attribute is required to have a value or not.
    :type required: bool

    :param changeable: Whether attribute value can be changed.
    :type changeable: bool

    :param deletable: Whether attribute value can be deleted.
    :type deletable: bool

    :param finalized: If True, attribute can't be overridden by subclasses.
    :type finalized: bool

    :param abstracted: If True, attribute needs to be overridden by subclasses.
    :type abstracted: bool

    :param qual_name: Optional type qualified name for the generated class.
    :type qual_name: str or None

    :param unique: Whether generated class should have a unique descriptor.
    :type unique: bool

    :param reactions: Reaction functions ordered by priority.
    :type reactions: str or collections.abc.Callable or None or \
collections.abc.Iterable[str or collections.abc.Callable]

    :param mutable: Whether generated class should have public mutable interface.
    :type mutable: bool

    :return: Set attribute.
    :rtype: objetto.objects.Attribute[objetto.objects.MutableSetObject or \
objetto.objects.SetObject]

    :raises TypeError: Invalid parameter type.
    :raises ValueError: Invalid parameter value.
    """

    # Get module from caller if not provided.
    module = get_caller_module() if module is None else module

    # Make set class.
    with ReraiseContext((TypeError, ValueError), "defining 'set_attribute'"):
        set_type = set_cls(
            types=types,
            subtypes=subtypes,
            checked=checked,
            module=module,
            factory=factory,
            serialized=serialized,
            serializer=serializer,
            deserializer=deserializer,
            represented=represented,
            child=child,
            history=history,
            data=data,
            custom_data_relationship=custom_data_relationship,
            qual_name=qual_name,
            unique=unique,
            reactions=reactions,
            mutable=mutable,
        )  # type: Union[Type[MutableSetObject[T]], Type[SetObject[T]]]

    # Factory for set object relationship.
    def set_factory(initial=(), app=None, **_):
        """Factory for the whole set object."""
        if type(initial) is set_type and initial.app is app:
            return initial
        else:
            return set_type(app, initial)

    # Get default value/factory.
    if changeable and not required and default is MISSING and default_factory is None:
        default = ()

    # Relationship.
    with ReraiseContext((TypeError, ValueError), "defining 'attribute'"):
        relationship = Relationship(
            types=set_type,
            subtypes=False,
            checked=False,
            module=module,
            factory=set_factory,
            serialized=serialized,
            serializer=None,
            deserializer=None,
            represented=represented,
            child=child,
            history=history,
            data=data,
            data_relationship=None,
        )

    # Make attribute.
    with ReraiseContext((TypeError, ValueError), "defining 'attribute'"):
        attribute_ = Attribute(
            relationship=relationship,
            default=default,
            default_factory=default_factory,
            module=module,
            required=required,
            changeable=changeable,
            deletable=deletable,
            finalized=finalized,
            abstracted=abstracted,
            delegated=False,
            dependencies=None,
            deserialize_to=None,
        )  # type: Attribute[MutableSetObject[T]]

    return attribute_


def protected_set_attribute_pair(
    types=(),  # type: Union[Type[VT], NT, str, Iterable[Union[Type[VT], NT, str]]]
    subtypes=False,  # type: bool
    checked=None,  # type: Optional[bool]
    module=None,  # type: Optional[str]
    factory=None,  # type: LazyFactory
    serialized=None,  # type: Optional[bool]
    serializer=None,  # type: LazyFactory
    deserializer=None,  # type: LazyFactory
    represented=True,  # type: bool
    child=True,  # type: bool
    history=None,  # type: Optional[bool]
    data=None,  # type: Optional[bool]
    custom_data_relationship=None,  # type: Optional[DataRelationship]
    default=MISSING,  # type: Any
    default_factory=None,  # type: LazyFactory
    finalized=False,  # type: bool
    abstracted=False,  # type: bool
    qual_name=None,  # type: Optional[str]
    unique=False,  # type: bool
    reactions=None,  # type: ReactionsType
):
    # type: (...) -> Tuple[ProxySetAttribute[T], SetAttribute[T]]
    """
    Make protected-public set attribute pair.

    :param types: Types.
    :type types: str or type or None or tuple[str or type or None]

    :param subtypes: Whether to accept subtypes.
    :type subtypes: bool

    :param checked: Whether to perform runtime type check.
    :type checked: bool

    :param module: Module path for lazy types/factories.
    :type module: str or None

    :param factory: Value factory.
    :type factory: str or collections.abc.Callable or None

    :param serialized: Whether should be serialized.
    :type serialized: bool

    :param serializer: Custom serializer.
    :type serializer: str or collections.abc.Callable or None

    :param deserializer: Custom deserializer.
    :type deserializer: str or collections.abc.Callable or None

    :param represented: Whether should be represented.
    :type represented: bool

    :param child: Whether object values should be adopted as children.
    :type child: bool

    :param history: Whether to propagate the history to the child object value.
    :type history: bool

    :param data: Whether to generate data for the value.
    :type data: bool

    :param custom_data_relationship: Custom data relationship.
    :type custom_data_relationship: objetto.data.DataRelationship or None

    :param default: Default value.

    :param default_factory: Default value factory.
    :type default_factory: str or collections.abc.Callable or None

    :param finalized: If True, attribute can't be overridden by subclasses.
    :type finalized: bool

    :param abstracted: If True, attribute needs to be overridden by subclasses.
    :type abstracted: bool

    :param qual_name: Optional type qualified name for the generated class.
    :type qual_name: str or None

    :param unique: Whether generated class should have a unique descriptor.
    :type unique: bool

    :param reactions: Reaction functions ordered by priority.
    :type reactions: str or collections.abc.Callable or None or \
collections.abc.Iterable[str or collections.abc.Callable]

    :return: Protected-public set attribute pair.
    :rtype: tuple[objetto.objects.Attribute[objetto.objects.ProxySetObject], \
objetto.objects.Attribute[objetto.objects.SetObject]]

    :raises TypeError: Invalid parameter type.
    :raises ValueError: Invalid parameter value.
    """

    # Get module from caller if not provided.
    module = get_caller_module() if module is None else module

    # Make public set attribute.
    if default is MISSING and default_factory is None:
        default = ()

    public_attribute = cast(
        "SetAttribute[T]",
        set_attribute(
            types=types,
            subtypes=subtypes,
            checked=checked,
            module=module,
            factory=factory,
            serialized=serialized,
            serializer=serializer,
            deserializer=deserializer,
            represented=represented,
            child=child,
            history=history,
            data=data,
            custom_data_relationship=custom_data_relationship,
            default=default,
            default_factory=default_factory,
            required=False,
            changeable=False,
            deletable=False,
            finalized=finalized,
            abstracted=abstracted,
            qual_name=qual_name,
            unique=unique,
            reactions=reactions,
            mutable=False,
        ),
    )

    # Make protected attribute.
    protected_attribute = cast(
        "ProxySetAttribute[T]",
        attribute(
            types=ProxySetObject,
            subtypes=False,
            checked=False,
            module=module,
            represented=False,
            child=False,
            required=False,
            finalized=finalized,
            abstracted=abstracted,
            delegated=True,
            dependencies=(public_attribute,),
        ),
    )

    def getter(iobj):
        """Getter delegate."""
        public_name = iobj.__.cls._attribute_names[public_attribute]
        return ProxySetObject(iobj[public_name])

    protected_attribute.getter(getter)

    return protected_attribute, public_attribute


def _prepare_reactions(reactions=None):
    # type: (ReactionsType) -> Dict[str, BaseReaction]
    """
    Conform reactions parameter value into a dictionary with reaction methods.

    :param reactions: Input reactions.
    :return: Dictionary with reaction methods.
    """
    dct = {}  # type: Dict[str, BaseReaction]
    if reactions is None:
        return dct
    if isinstance(reactions, string_types) or not isinstance(
        reactions, collections_abc.Iterable
    ):
        reactions = (reactions,)
    for i, reaction_ in enumerate(reactions):
        if isinstance(reaction_, BaseReaction) and reaction_.priority != i:
            reaction_ = reaction_.set_priority(i)
        elif reaction_ is None:
            continue
        else:
            reaction_decorator = reaction(priority=i)  # type: ignore
            # noinspection PyArgumentList
            reaction_ = reaction_decorator(import_factory(reaction_))  # type: ignore
        dct["__reaction{}".format(i)] = cast("BaseReaction", reaction_)
    return dct


def dict_cls(
    types=(),  # type: Union[Type[VT], NT, str, Iterable[Union[Type[VT], NT, str]]]
    subtypes=False,  # type: bool
    checked=None,  # type: Optional[bool]
    module=None,  # type: Optional[str]
    factory=None,  # type: LazyFactory
    serialized=None,  # type: Optional[bool]
    serializer=None,  # type: LazyFactory
    deserializer=None,  # type: LazyFactory
    represented=True,  # type: bool
    key_types=(),  # type: Union[Type[KT], NT, str, Iterable[Union[Type[KT], NT, str]]]
    key_subtypes=False,  # type: bool
    key_factory=None,  # type: LazyFactory
    child=True,  # type: bool
    history=None,  # type: Optional[bool]
    data=None,  # type: Optional[bool]
    custom_data_relationship=None,  # type: Optional[DataRelationship]
    qual_name=None,  # type: Optional[str]
    unique=False,  # type: bool
    reactions=None,  # type: ReactionsType
    mutable=True,  # type: bool
):
    # type: (...) -> Union[Type[MutableDictObject[KT, VT]], Type[DictObject[KT, VT]]]
    """
    Make auxiliary dictionary object class.

    :param types: Types.
    :type types: str or type or None or tuple[str or type or None]

    :param subtypes: Whether to accept subtypes.
    :type subtypes: bool

    :param checked: Whether to perform runtime type check.
    :type checked: bool

    :param module: Module path for lazy types/factories.
    :type module: str or None

    :param factory: Value factory.
    :type factory: str or collections.abc.Callable or None

    :param serialized: Whether should be serialized.
    :type serialized: bool

    :param serializer: Custom serializer.
    :type serializer: str or collections.abc.Callable or None

    :param deserializer: Custom deserializer.
    :type deserializer: str or collections.abc.Callable or None

    :param represented: Whether should be represented.
    :type represented: bool

    :param child: Whether object values should be adopted as children.
    :type child: bool

    :param history: Whether to propagate the history to the child object value.
    :type history: bool

    :param data: Whether to generate data for the value.
    :type data: bool

    :param custom_data_relationship: Custom data relationship.
    :type custom_data_relationship: objetto.data.DataRelationship or None

    :param key_types: Key types.
    :type key_types: str or type or None or tuple[str or type or None]

    :param key_subtypes: Whether to accept subtypes for the keys.
    :type key_subtypes: bool

    :param key_factory: Key factory.
    :type key_factory: str or collections.abc.Callable or None

    :param qual_name: Optional type qualified name for the generated class.
    :type qual_name: str or None

    :param unique: Whether generated class should have a unique descriptor.
    :type unique: bool

    :param reactions: Reaction functions ordered by priority.
    :type reactions: str or collections.abc.Callable or None or \
collections.abc.Iterable[str or collections.abc.Callable]

    :param mutable: Whether generated class should have public mutable interface.
    :type mutable: bool

    :return: Dictionary object class.
    :rtype: type[objetto.objects.DictObject or objetto.objects.MutableDictObject]

    :raises TypeError: Invalid parameter type.
    :raises ValueError: Invalid parameter value.
    """

    # Get module from caller if not provided.
    module = get_caller_module() if module is None else module

    # Relationship.
    with ReraiseContext((TypeError, ValueError), "defining 'dict_cls'"):
        relationship = Relationship(
            types=types,
            subtypes=subtypes,
            checked=checked,
            module=module,
            factory=factory,
            serialized=serialized,
            serializer=serializer,
            deserializer=deserializer,
            represented=represented,
            child=child,
            history=history,
            data=data,
            data_relationship=custom_data_relationship,
        )

    # Key relationship.
    with ReraiseContext((TypeError, ValueError), "defining 'dict_cls'"):
        key_relationship = KeyRelationship(
            types=key_types,
            subtypes=key_subtypes,
            checked=checked,
            module=module,
            factory=key_factory,
        )
        dct = {"_key_relationship": key_relationship}  # type: Dict[str, Any]

    # Reactions.
    with ReraiseContext((TypeError, ValueError), "defining 'dict_cls'"):
        dct.update(_prepare_reactions(reactions))

    # Make class.
    cls_kwargs = dict(
        relationship=relationship,
        qual_name=qual_name,
        module=module,
        unique_descriptor_name="unique_hash" if unique else None,
        dct=dct,
    )  # type: Dict[str, Any]
    with ReraiseContext(TypeError, "defining 'dict_cls'"):
        if mutable:
            mutable_base = MutableDictObject  # type: Type[MutableDictObject[KT, VT]]
            return make_auxiliary_cls(mutable_base, **cls_kwargs)
        else:
            base = DictObject  # type: Type[DictObject[KT, VT]]
            return make_auxiliary_cls(base, **cls_kwargs)


def list_cls(
    types=(),  # type: Union[Type[T], NT, str, Iterable[Union[Type[T], NT, str]]]
    subtypes=False,  # type: bool
    checked=None,  # type: Optional[bool]
    module=None,  # type: Optional[str]
    factory=None,  # type: LazyFactory
    serialized=None,  # type: Optional[bool]
    serializer=None,  # type: LazyFactory
    deserializer=None,  # type: LazyFactory
    represented=True,  # type: bool
    child=True,  # type: bool
    history=None,  # type: Optional[bool]
    data=None,  # type: Optional[bool]
    custom_data_relationship=None,  # type: Optional[DataRelationship]
    qual_name=None,  # type: Optional[str]
    unique=False,  # type: bool
    reactions=None,  # type: ReactionsType
    mutable=True,  # type: bool
):
    # type: (...) -> Union[Type[MutableListObject[T]], Type[ListObject[T]]]
    """
    Make auxiliary list object class.

    :param types: Types.
    :type types: str or type or None or tuple[str or type or None]

    :param subtypes: Whether to accept subtypes.
    :type subtypes: bool

    :param checked: Whether to perform runtime type check.
    :type checked: bool

    :param module: Module path for lazy types/factories.
    :type module: str or None

    :param factory: Value factory.
    :type factory: str or collections.abc.Callable or None

    :param serialized: Whether should be serialized.
    :type serialized: bool

    :param serializer: Custom serializer.
    :type serializer: str or collections.abc.Callable or None

    :param deserializer: Custom deserializer.
    :type deserializer: str or collections.abc.Callable or None

    :param represented: Whether should be represented.
    :type represented: bool

    :param child: Whether object values should be adopted as children.
    :type child: bool

    :param history: Whether to propagate the history to the child object value.
    :type history: bool

    :param data: Whether to generate data for the value.
    :type data: bool

    :param custom_data_relationship: Custom data relationship.
    :type custom_data_relationship: objetto.data.DataRelationship or None

    :param qual_name: Optional type qualified name for the generated class.
    :type qual_name: str or None

    :param unique: Whether generated class should have a unique descriptor.
    :type unique: bool

    :param reactions: Reaction functions ordered by priority.
    :type reactions: str or collections.abc.Callable or None or \
collections.abc.Iterable[str or collections.abc.Callable]

    :param mutable: Whether generated class should have public mutable interface.
    :type mutable: bool

    :return: List object class.
    :rtype: type[objetto.objects.ListObject or objetto.objects.MutableListObject]

    :raises TypeError: Invalid parameter type.
    :raises ValueError: Invalid parameter value.
    """

    # Get module from caller if not provided.
    module = get_caller_module() if module is None else module

    # Relationship.
    with ReraiseContext((TypeError, ValueError), "defining 'list_cls'"):
        relationship = Relationship(
            types=types,
            subtypes=subtypes,
            checked=checked,
            module=module,
            factory=factory,
            serialized=serialized,
            serializer=serializer,
            deserializer=deserializer,
            represented=represented,
            child=child,
            history=history,
            data=data,
            data_relationship=custom_data_relationship,
        )

    # Reactions.
    with ReraiseContext((TypeError, ValueError), "defining 'list_cls'"):
        dct = _prepare_reactions(reactions)

    # Make class.
    cls_kwargs = dict(
        relationship=relationship,
        qual_name=qual_name,
        module=module,
        unique_descriptor_name="unique_hash" if unique else None,
        dct=dct,
    )  # type: Dict[str, Any]
    with ReraiseContext(TypeError, "defining 'list_cls'"):
        if mutable:
            mutable_base = MutableListObject  # type: Type[MutableListObject[T]]
            return make_auxiliary_cls(mutable_base, **cls_kwargs)
        else:
            base = ListObject  # type: Type[ListObject[T]]
            return make_auxiliary_cls(base, **cls_kwargs)


def set_cls(
    types=(),  # type: Union[Type[T], NT, str, Iterable[Union[Type[T], NT, str]]]
    subtypes=False,  # type: bool
    checked=None,  # type: Optional[bool]
    module=None,  # type: Optional[str]
    factory=None,  # type: LazyFactory
    serialized=None,  # type: Optional[bool]
    serializer=None,  # type: LazyFactory
    deserializer=None,  # type: LazyFactory
    represented=True,  # type: bool
    child=True,  # type: bool
    history=None,  # type: Optional[bool]
    data=None,  # type: Optional[bool]
    custom_data_relationship=None,  # type: Optional[DataRelationship]
    qual_name=None,  # type: Optional[str]
    unique=False,  # type: bool
    reactions=None,  # type: ReactionsType
    mutable=True,  # type: bool
):
    # type: (...) -> Union[Type[MutableSetObject[T]], Type[SetObject[T]]]
    """
    Make auxiliary set object class.

    :param types: Types.
    :type types: str or type or None or tuple[str or type or None]

    :param subtypes: Whether to accept subtypes.
    :type subtypes: bool

    :param checked: Whether to perform runtime type check.
    :type checked: bool

    :param module: Module path for lazy types/factories.
    :type module: str or None

    :param factory: Value factory.
    :type factory: str or collections.abc.Callable or None

    :param serialized: Whether should be serialized.
    :type serialized: bool

    :param serializer: Custom serializer.
    :type serializer: str or collections.abc.Callable or None

    :param deserializer: Custom deserializer.
    :type deserializer: str or collections.abc.Callable or None

    :param represented: Whether should be represented.
    :type represented: bool

    :param child: Whether object values should be adopted as children.
    :type child: bool

    :param history: Whether to propagate the history to the child object value.
    :type history: bool

    :param data: Whether to generate data for the value.
    :type data: bool

    :param custom_data_relationship: Custom data relationship.
    :type custom_data_relationship: objetto.data.DataRelationship or None

    :param qual_name: Optional type qualified name for the generated class.
    :type qual_name: str or None

    :param unique: Whether generated class should have a unique descriptor.
    :type unique: bool

    :param reactions: Reaction functions ordered by priority.
    :type reactions: str or collections.abc.Callable or None or \
collections.abc.Iterable[str or collections.abc.Callable]

    :param mutable: Whether generated class should have public mutable interface.
    :type mutable: bool

    :return: Set object class.
    :rtype: type[objetto.objects.SetObject or objetto.objects.MutableSetObject]

    :raises TypeError: Invalid parameter type.
    :raises ValueError: Invalid parameter value.
    """

    # Get module from caller if not provided.
    module = get_caller_module() if module is None else module

    # Relationship.
    with ReraiseContext((TypeError, ValueError), "defining 'set_cls'"):
        relationship = Relationship(
            types=types,
            subtypes=subtypes,
            checked=checked,
            module=module,
            factory=factory,
            serialized=serialized,
            serializer=serializer,
            deserializer=deserializer,
            represented=represented,
            child=child,
            history=history,
            data=data,
            data_relationship=custom_data_relationship,
        )

    # Reactions.
    with ReraiseContext((TypeError, ValueError), "defining 'set_cls'"):
        dct = _prepare_reactions(reactions)

    # Make class.
    cls_kwargs = dict(
        relationship=relationship,
        qual_name=qual_name,
        module=module,
        unique_descriptor_name="unique_hash" if unique else None,
        dct=dct,
    )  # type: Dict[str, Any]
    with ReraiseContext(TypeError, "defining 'set_cls'"):
        if mutable:
            mutable_base = MutableSetObject  # type: Type[MutableSetObject[T]]
            return make_auxiliary_cls(mutable_base, **cls_kwargs)
        else:
            base = SetObject  # type: Type[SetObject[T]]
            return make_auxiliary_cls(base, **cls_kwargs)



# -*- coding: utf-8 -*-
"""Base classes and metaclasses."""

from abc import abstractmethod
from contextlib import contextmanager
from inspect import getmro
from typing import TYPE_CHECKING, Callable, Generic, Type, TypeVar, cast

try:
    from types import new_class
except ImportError:
    new_class = None  # type: ignore

try:
    from typing import final
except ImportError:
    final = lambda f: f  # type: ignore

from uuid import uuid4
from weakref import WeakKeyDictionary, WeakValueDictionary

try:
    import collections.abc as collections_abc
except ImportError:
    import collections as collections_abc  # type: ignore

from decorator import decorator
from qualname import qualname  # type: ignore
from six import iteritems, with_metaclass
from slotted import (
    SlottedABC,
    SlottedABCMeta,
    SlottedContainer,
    SlottedHashable,
    SlottedIterable,
    SlottedSized,
)

if TYPE_CHECKING:
    from typing import (
        Any,
        Dict,
        FrozenSet,
        Iterable,
        Iterator,
        List,
        Mapping,
        MutableMapping,
        Optional,
        Set,
        Union,
    )

__all__ = [
    "MISSING",
    "ABSTRACT_TAG",
    "FINAL_CLASS_TAG",
    "FINAL_METHOD_TAG",
    "INITIALIZING_TAG",
    "final",
    "init_context",
    "init",
    "simplify_member_names",
    "make_base_cls",
    "BaseMeta",
    "Base",
    "abstract_member",
    "Generic",
    "BaseHashable",
    "BaseSized",
    "BaseIterable",
    "BaseContainer",
    "BaseCollection",
    "BaseProtectedCollection",
    "BaseInteractiveCollection",
    "BaseMutableCollection",
]

# noinspection PyTypeChecker
F = TypeVar("F", Callable, Type)  # Callable type.
T = TypeVar("T")  # Any type.
KT = TypeVar("KT")  # Key type.
VT = TypeVar("VT")  # Value type.
T_co = TypeVar("T_co", covariant=True)  # Any type covariant containers.
VT_co = TypeVar("VT_co", covariant=True)  # Value type covariant containers.

MISSING = object()
ABSTRACT_TAG = "__isabstractmethod__"
FINAL_CLASS_TAG = "__isfinalclass__"
FINAL_METHOD_TAG = "__isfinalmethod__"
INITIALIZING_TAG = "__isinitializing__"

__final = final
__base_cls_cache = WeakValueDictionary()  # type: MutableMapping[str, Type[Base]]


class _GenericMeta(SlottedABCMeta):
    """Workaround for 'Generic' class having a metaclass in older versions of typing."""

    def __getitem__(cls, _):
        return cls


type.__setattr__(_GenericMeta, "__name__", "SlottedABCMeta")
if hasattr(SlottedABCMeta, "__qualname__"):
    type.__setattr__(_GenericMeta, "__qualname__", "SlottedABCMeta")
type.__setattr__(_GenericMeta, "__module__", SlottedABCMeta.__module__)
try:
    type.__setattr__(_GenericMeta, "__doc__", SlottedABCMeta.__doc__)
except AttributeError:
    pass


class _Generic(with_metaclass(_GenericMeta, SlottedABC)):
    pass


type.__setattr__(_Generic, "__name__", "Generic")
if hasattr(Generic, "__qualname__"):
    type.__setattr__(_Generic, "__qualname__", "Generic")
type.__setattr__(_Generic, "__module__", Generic.__module__)
try:
    type.__setattr__(_Generic, "__doc__", Generic.__doc__)
except AttributeError:
    pass

globals()["SlottedABCMeta"] = _GenericMeta
globals()["Generic"] = _Generic


def _final(obj):
    # type: (F) -> F
    """
    Decorator based on :func:`typing.final` that enables runtime checking for
    :class:`objetto.bases.Base` classes.

    .. code:: python

        >>> from objetto.bases import Base, final

        >>> @final
        ... class FinalClass(Base):  # final class
        ...     pass
        ...
        >>> class Class(Base):
        ...     @final
        ...     def final_method(self):  # final method
        ...         pass
        ...

    :param obj: Method function or class.
    :type obj: function or type

    :return: Decorated method function or class.
    :rtype: function or type
    """
    if isinstance(obj, type):
        type.__setattr__(obj, FINAL_CLASS_TAG, True)
        return __final(obj)
    else:

        @decorator
        def final_(obj_, *args, **kwargs):
            """Decorator for final methods."""
            return obj_(*args, **kwargs)

        decorated = final_(obj)
        object.__setattr__(decorated, FINAL_METHOD_TAG, True)
        return __final(decorated)


# Replace typing.final with our custom decorator for runtime checking.
globals()["final"] = _final


def simplify_member_names(names):
    # type: (Iterable[str]) -> Iterator[str]
    """
    Iterate over member names and only yield the simplified ones.

    :param names: Input names.
    :type names: str

    :return: Simplified names iterator.
    :rtype: collections.abc.Iterator[str]
    """
    return (n for n in names if not ("__" in n and n.startswith("_")))


@contextmanager
def init_context(obj):
    # type: (Base) -> Iterator
    """
    Context manager that sets the initializing tag for :class:`objetto.bases.Base`
    objects.

    .. code:: python

        >>> from objetto.bases import init_context

        >>> class MyClass(Base):
        ...     def __init__(self):  # initialization tag is implicitly set
        ...         print(("__init__", self._initializing))
        ...
        ...     def init(self):
        ...         with init_context(self):
        ...             print(("init (inside of context)", self._initializing))
        ...         print(("init (outside of context)", self._initializing))
        ...
        >>> my_obj = MyClass()
        ('__init__', True)
        >>> my_obj.init()
        ('init (inside of context)', True)
        ('init (outside of context)', False)

    :param obj: Instance of :class:`objetto.bases.Base`.
    :type obj: objetto.bases.Base

    :return: Context manager.
    :rtype: contextlib.AbstractContextManager
    """
    previous = getattr(obj, INITIALIZING_TAG, False)
    # noinspection PyCallByClass
    object.__setattr__(obj, INITIALIZING_TAG, True)
    try:
        yield
    finally:
        # noinspection PyCallByClass
        object.__setattr__(obj, INITIALIZING_TAG, previous)


@decorator
def init(func, *args, **kwargs):
    # type: (F, Any, Any) -> F
    """
    Method decorator that sets the initializing tag for :class:`objetto.bases.Base`
    objects.

    .. code:: python

        >>> from objetto.bases import init

        >>> class MyClass(Base):
        ...     def __init__(self):  # initialization tag is implicitly set
        ...         print(("__init__", self._initializing))
        ...
        ...     @init
        ...     def init(self):
        ...         print(("init", self._initializing))
        ...
        ...     def not_init(self):
        ...         print(("not_init", self._initializing))
        ...
        >>> my_obj = MyClass()
        ('__init__', True)
        >>> my_obj.init()
        ('init', True)
        >>> my_obj.not_init()
        ('not_init', False)

    :param func: Method function.
    :type func: function

    :return: Decorated method function.
    :rtype: function
    """
    self = args[0]
    with init_context(self):
        result = func(*args, **kwargs)
    return result


# noinspection PyTypeChecker
_B = TypeVar("_B", bound="Base")


def _make_base_cls(
    base=None,  # type: Optional[Type[_B]]
    qual_name=None,  # type: Optional[str]
    module=None,  # type: Optional[str]
    dct=None,  # type: Optional[Mapping[str, Any]]
    uuid=None,  # type: Optional[str]
):
    # type: (...) -> Type[_B]
    """
    Make a subclass of :class:`objetto.bases.Base` on the fly.

    :param base: Base class.
    :param qual_name: Qualified name.
    :param module: Module.
    :param dct: Members dictionary.
    :param uuid: UUID used for caching.
    :return: Generated subclass.
    """

    # Get base.
    if base is None:
        _base = Base
    else:
        _base = cast("Type[_B]", base)

    # Get name.
    qual_name = qual_name or _base.__fullname__ or _base.__name__ or ""
    name = qual_name.split(".")[-1]

    # Get module.
    module = module or _base.__module__

    # Get metaclass and uuid.
    mcs = type(_base)
    uuid = str(uuid4()) if uuid is None else uuid

    # Copy dct.
    dct_copy = dict(dct or {})  # type: Dict[str, Any]

    # Define reduce method for pickling instances of the subclass.
    def __reduce__(self):
        state = self.__getstate__()
        if type(self).__dict__.get("__base_cls_uuid__", None) == uuid:
            return _make_base_instance, (
                _base,
                qual_name,
                module,
                state,
                dct_copy,
                uuid,
            )
        else:
            return _make_base_subclass_instance, (type(self), state)

    # Fallback qualified name property for python 2.7.
    if not hasattr(object, "__qualname__"):

        class QualNameClsProperty(object):
            __slots__ = ()

            def __get__(self, instance, owner):
                if owner is None:
                    return self
                if instance is not None:
                    owner = type(instance)
                if owner.__dict__.get("__qualname__", None) is not self:
                    error = "type object '{}' has no attribute '__qualname__'".format(
                        owner.__name__,
                    )
                    raise AttributeError(error)
                return qual_name

        __qualname__ = QualNameClsProperty()  # type: Union[str, QualNameClsProperty]
    else:
        __qualname__ = qual_name

    # Assemble class dict by copying dct once again.
    cls_dct = dct_copy.copy()  # type: Dict[str, Any]
    cls_dct_update = {
        "__base_cls_uuid__": uuid,
        "__reduce__": __reduce__,
        "__qualname__": __qualname__,
        "__module__": module,
    }  # type: Dict[str, Any]
    cls_dct.update(cls_dct_update)

    # Make new subclass and cache it by UUID.
    if new_class is not None:

        def exec_body(ns):
            for k, v in iteritems(cls_dct):
                ns[k] = v
            return ns

        cls = cast(
            "Type[_B]", new_class(name, (_base,), {"metaclass": BaseMeta}, exec_body)
        )
    else:
        cls = mcs(name, (_base,), cls_dct)

    __base_cls_cache[uuid] = cls

    return cls


def _make_base_subclass_instance(
    cls,  # type: Type[_B]
    state,  # type: Dict[str, Any]
):
    # type: (...) -> _B
    """
    Make an instance of a subclass of a generated :class:`objetto.bases.Base`.

    :param cls: Base subclass.
    :param state: Pickled state.
    :return: Generated instance.
    """
    self = cast("_B", cls.__new__(cls))
    self.__setstate__(state or {})
    return self


def _make_base_instance(
    base=None,  # type: Optional[Type[_B]]
    qual_name=None,  # type: Optional[str]
    module=None,  # type: Optional[str]
    state=None,  # type: Optional[Dict[str, Any]]
    dct=None,  # type: Optional[Mapping[str, Any]]
    uuid=None,  # type: Optional[str]
):
    # type: (...) -> _B
    """
    Make an instance of a subclass of :class:`objetto.bases.Base` on the fly.

    :param base: Base class.
    :param qual_name: Qualified name.
    :param module: Module.
    :param state: Pickled state.
    :param dct: Members dictionary.
    :param uuid: UUID used for caching.
    :return: Generated instance.
    """

    # Try to get class from cache using UUID.
    try:
        if uuid is None:
            raise KeyError()
        cls = __base_cls_cache[uuid]

    # Not cached, make a new subclass and use it.
    except KeyError:
        cls = _make_base_cls(
            base=base,
            qual_name=qual_name,
            module=module,
            dct=dct,
            uuid=uuid,
        )

    # Make new instance and unpickle its state.
    self = cast("_B", cls.__new__(cls))
    self.__setstate__(state or {})

    return self


def make_base_cls(
    base=None,  # type: Optional[Type[_B]]
    qual_name=None,  # type: Optional[str]
    module=None,  # type: Optional[str]
    dct=None,  # type: Optional[Mapping[str, Any]]
):
    # type: (...) -> Type[_B]
    """
    Make a subclass of :class:`objetto.bases.Base` on the fly.

    :param base: Base class.
    :type base: type[objetto.bases.Base]

    :param qual_name: Qualified name.
    :type qual_name: str or None

    :param module: Module.
    :type module: str or None

    :param dct: Members dictionary.
    :param dct: dict[str, Any] or None

    :return: Generated subclass.
    :rtype: type[objetto.bases.Base]
    """
    return _make_base_cls(base, qual_name, module, dct=dct)


class BaseMeta(SlottedABCMeta):
    """
    Metaclass for :class:`objetto.bases.Base`.

    Features:
      - Forces the use of `__slots__`.
      - Forces `__hash__` to be declared if `__eq__` was declared.
      - Decorates `__init__` methods to update the initializing tag.
      - Prevents base class attributes from changing.
      - Runtime checking for `final` decorated classes/methods.
      - Implements `__fullname__` class property for backporting qualified name.
    """

    __open_attributes = WeakKeyDictionary(
        {}
    )  # type: MutableMapping[Type, FrozenSet[str]]

    @staticmethod
    def __new__(mcs, name, bases, dct):
        """Make `BaseMeta` class."""
        dct = dict(dct)

        # Force '__hash__' to be declared if '__eq__' is declared.
        if "__eq__" in dct and "__hash__" not in dct:
            error = "declared '__eq__' in '{}', but didn't declare '__hash__'".format(
                name
            )
            raise TypeError(error)

        # Always decorate '__init__' method with 'init'.
        if "__init__" in dct:
            dct["__init__"] = init(dct["__init__"])

        return super(BaseMeta, mcs).__new__(mcs, name, bases, dct)

    def __init__(cls, name, bases, dct):
        super(BaseMeta, cls).__init__(name, bases, dct)

        # Traverse the MRO from the bottom.
        final_method_names = set()  # type: Set[str]
        final_cls = None  # type: Optional[Type]
        required_bases = list(reversed(getmro(SlottedABC)))  # type: List[Type]
        seen_bases = []  # type: List[Type]
        open_attributes = set()  # type: Set[str]
        for base in reversed(getmro(cls)):

            # Check bases for valid resolution order.
            expected_base = seen_bases[-1] if seen_bases else object
            if not issubclass(base, expected_base):
                error = (
                    "invalid resolution order when defining '{}'; "
                    "base '{}' does not inherit from '{}'"
                ).format(name, base.__name__, expected_base.__name__)
                raise TypeError(error)
            elif required_bases and base is required_bases[-1]:
                seen_bases.append(required_bases.pop())

            # Get open class attributes if non-Base, start requiring Base otherwise.
            if not isinstance(base, BaseMeta):
                open_attributes.update(base.__dict__)
            elif not required_bases:
                required_bases.append(base)

            # Prevent subclassing final classes.
            if getattr(base, FINAL_CLASS_TAG, False) is True:
                if final_cls is not None:
                    error = "can't subclass final class '{}'".format(final_cls.__name__)
                    raise TypeError(error)
                final_cls = base

            # Prevent overriding of final methods.
            for member_name, member in iteritems(base.__dict__):
                if member_name in final_method_names:
                    error = "can't override final member '{}'".format(member_name)
                    raise TypeError(error)

                if isinstance(member, property):
                    is_final = getattr(member.fget, FINAL_METHOD_TAG, False) is True
                elif isinstance(member, (staticmethod, classmethod)):
                    is_final = getattr(member.__func__, FINAL_METHOD_TAG, False) is True
                else:
                    is_final = getattr(member, FINAL_METHOD_TAG, False) is True

                if is_final:
                    final_method_names.add(member_name)

        # Store open class attributes.
        type(cls).__open_attributes[cls] = frozenset(open_attributes)

    def __repr__(cls):
        # type: () -> str
        """
        Get representation.

        :return: Representation.
        """
        module = cls.__module__
        name = cls.__fullname__
        return "<class{space}{quote}{module}{name}{quote}>".format(
            space=" " if module or name else "",
            quote="'" if module or name else "",
            module=module + "." if module else "",
            name=name if name else "",
        )

    def __dir__(cls):
        # type: () -> List[str]
        """
        Get a simplified list of member names.

        :return: List of member names.
        """
        member_names = set()  # type: Set[str]
        for base in reversed(getmro(type(cls))):
            if base is object or base is type:
                continue
            member_names.update(simplify_member_names(base.__dict__))
        for base in reversed(getmro(cls)):
            if base is object or base is type:
                continue
            member_names.update(simplify_member_names(base.__dict__))
        return sorted(member_names)

    @final
    def __setattr__(cls, name, value):
        # type: (str, Any) -> None
        """
        Set class attribute.

        :param name: Name.
        :param value: Value.
        :raises AttributeError: Read-only attribute.
        """
        open_attributes = type(cls).__open_attributes.get(cls)
        if open_attributes is not None and name not in open_attributes:
            error = "class attribute '{}' is read-only".format(name)
            raise AttributeError(error)
        super(BaseMeta, cls).__setattr__(name, value)

    @final
    def __delattr__(cls, name):
        # type: (str) -> None
        """
        Delete class attribute.

        :param name: Name.
        :raises AttributeError: Read-only attribute.
        """
        open_attributes = type(cls).__open_attributes.get(cls)
        if open_attributes is not None and name not in open_attributes:
            error = "class attribute '{}' is read-only".format(name)
            raise AttributeError(error)
        super(BaseMeta, cls).__delattr__(name)

    @property
    @final
    def __fullname__(cls):
        # type: () -> str
        """
        Get qualified class name if possible, fall back to class name otherwise.

        :return: Full class name.
        """
        try:
            name = qualname(cls)
            if not name:
                raise AttributeError()
        except AttributeError:
            name = cls.__name__
        return name


class Base(with_metaclass(BaseMeta, SlottedABC)):
    """
    Base class for all `Objetto` types.

    Inherits from:
      - :class:`slotted.SlottedABC`

    Inherited By:
      - :class:`objetto.bases.BaseHashable`
      - :class:`objetto.bases.BaseSized`
      - :class:`objetto.bases.BaseIterable`
      - :class:`objetto.bases.BaseContainer`
      - :class:`objetto.bases.BaseReaction`
      - :class:`objetto.bases.BaseFactory`
      - :class:`objetto.objects.UniqueDescriptor`

    Features:
      - Forces the use of `__slots__`.
      - Forces `__hash__` to be declared if `__eq__` was declared.
      - Property that tells whether instance is initializing or not.
      - Default implementation of `__copy__` raises an error.
      - Default implementation of `__ne__` returns the opposite of `__eq__`.
      - Prevents base class attributes from changing.
      - Runtime checking for `final` decorated classes/methods.
      - Simplified `__dir__` result that shows only relevant members for client code.
      - Implements `__fullname__` class property for backporting qualified name.
    """

    __slots__ = (INITIALIZING_TAG,)

    def __copy__(self):
        # type: () -> Any
        """
        Prevents shallow copy by default.

        :raises RuntimeError: Always raised.
        """
        error = "'{}' object can't be shallow copied".format(type(self).__fullname__)
        raise RuntimeError(error)

    def __repr__(self):
        # type: () -> str
        """
        Get representation using class' full name if possible.

        :return: Representation.
        :rtype: str
        """
        return "<{} at {}>".format(type(self).__fullname__, hex(id(self)))

    def __ne__(self, other):
        # type: (object) -> bool
        """
        Compare for inequality by negating the result of `__eq__`.
        This is a backport of the default python 3 behavior to python 2.

        :param other: Another object.

        :return: True if not equal.
        :rtype: bool or NotImplemented
        """
        result = self.__eq__(other)
        if result is NotImplemented:
            return NotImplemented
        return not result

    def __dir__(self):
        # type: () -> List[str]
        """
        Get a simplified list of member names.

        :return: List of member names.
        :rtype: list[str]
        """
        member_names = set()  # type: Set[str]
        for base in reversed(getmro(type(self))):
            if base is object or base is type:
                continue
            member_names.update(simplify_member_names(base.__dict__))
        return sorted(member_names)

    @property
    @final
    def _initializing(self):
        # type: () -> bool
        """
        Whether `__init__` (or any other initialization method) is running.

        :rtype: bool
        """
        return getattr(self, INITIALIZING_TAG, False)


@final
class AbstractMemberMeta(BaseMeta):
    """Metaclass for `AbstractMember`."""

    @staticmethod
    def __new__(mcs, name, bases, dct):
        """Make `AbstractMember` class."""
        dct[ABSTRACT_TAG] = True
        return super(AbstractMemberMeta, mcs).__new__(mcs, name, bases, dct)


@final
class AbstractMember(with_metaclass(AbstractMemberMeta, Base)):
    """Abstract member for classes."""

    __slots__ = ()

    @staticmethod
    def __new__(cls, *args, **kwargs):
        """
        Prevent instantiation.

        :raises TypeError: Always raised.
        """
        error = "'{}' can't be instantiated".format(cls.__name__)
        raise TypeError(error)

    def __repr__(self):
        # type: () -> str
        """
        Get representation.

        :return: Representation.
        """
        return "<abstract>"

    def __str__(self):
        # type: () -> str
        """
        Get string representation.

        :return: String representation.
        """
        return self.__repr__()


def abstract_member():
    # type: () -> Type[AbstractMember]
    """
    Used to indicate an abstract attribute member in a class.

    .. code:: python

        >>> from objetto.bases import Base, abstract_member

        >>> class AbstractClass(Base):
        ...     some_attribute = abstract_member()  # will prevent instatiation
        ...
        >>> obj = AbstractClass()
        Traceback (most recent call last):
        TypeError: Can't instantiate abstract class AbstractClass with abstract \
method...

        >>> class ConcreteClass(AbstractClass):
        ...     some_attribute = (1, 2, 3)  # concrete
        >>> obj = ConcreteClass()

    :return: Abstract member.
    """
    return AbstractMember


class BaseHashable(Base, SlottedHashable):
    """
    Base hashable.

    Inherits from:
      - :class:`objetto.bases.Base`
      - :class:`slotted.SlottedHashable`

    Inherited By:
      - :class:`objetto.bases.BaseState`
      - :class:`objetto.bases.BaseStructure`
      - :class:`objetto.bases.BaseRelationship`
      - :class:`objetto.bases.KeyRelationship`
      - :class:`objetto.bases.BaseAttribute`
      - :class:`objetto.history.HistoryDescriptor`

    Features:
      - Forces implementation of `__hash__` method.
    """

    __slots__ = ()

    @abstractmethod
    def __hash__(self):
        # type: () -> int
        """
        Get hash.

        :return: Hash.
        :rtype: int

        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()


class BaseSized(Base, SlottedSized):
    """
    Base sized.

    Inherits from:
      - :class:`objetto.bases.Base`
      - :class:`slotted.SlottedSized`

    Inherited By:
      - :class:`objetto.bases.BaseCollection`

    Features:
      - Has a length (count).
    """

    __slots__ = ()

    @abstractmethod
    def __len__(self):
        # type: () -> int
        """
        Get count.

        :return: Count.
        :rtype: int

        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()


class BaseIterable(Base, SlottedIterable, Generic[T_co]):
    """
    Base iterable.

    Inherits from:
      - :class:`objetto.bases.Base`
      - :class:`slotted.SlottedIterable`
      - :class:`typing.Generic`

    Inherited By:
      - :class:`objetto.bases.BaseCollection`

    Features:
      - Can be iterated over.
    """

    __slots__ = ()

    @abstractmethod
    def __iter__(self):
        # type: () -> Iterator
        """
        Iterate over.

        :return: Iterator.
        :rtype: collections.abc.Iterator

        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()


class BaseContainer(Base, SlottedContainer, Generic[T_co]):
    """
    Base container.

    Inherits from:
      - :class:`objetto.bases.Base`
      - :class:`slotted.SlottedContainer`
      - :class:`typing.Generic`

    Inherited By:
      - :class:`objetto.bases.BaseCollection`

    Features:
      - Contains values.
    """

    __slots__ = ()

    @abstractmethod
    def __contains__(self, content):
        # type: (Any) -> bool
        """
        Get whether content is present.

        :param content: Content.

        :return: True if contains.
        :rtype: bool

        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()


class BaseCollection(BaseSized, BaseIterable[T_co], BaseContainer[T_co]):
    """
    Base collection.

    Inherits from:
      - :class:`objetto.bases.BaseSized`
      - :class:`objetto.bases.BaseIterable`
      - :class:`objetto.bases.BaseContainer`

    Inherited By:
      - :class:`objetto.bases.BaseProtectedCollection`

    Features:
      - Method to find content that matches attribute values.
    """

    __slots__ = ()

    @abstractmethod
    def find_with_attributes(self, **attributes):
        # type: (Any) -> Any
        """
        Find first value that matches unique attribute values.

        :param attributes: Attributes to match.

        :return: Value that has matching attributes.

        :raises ValueError: No attributes provided or no match found.
        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()


# noinspection PyTypeChecker
_BPC = TypeVar("_BPC", bound="BaseProtectedCollection")


class BaseProtectedCollection(BaseCollection[T]):
    """
    Base protected collection.

    Inherits from:
      - :class:`objetto.bases.BaseCollection`

    Inherited By:
      - :class:`objetto.bases.BaseInteractiveCollection`
      - :class:`objetto.bases.BaseMutableCollection`
      - :class:`objetto.bases.BaseProtectedDict`
      - :class:`objetto.bases.BaseProtectedList`
      - :class:`objetto.bases.BaseProtectedSet`

    Features:
      - Has protected transformation methods.
      - Transformations return a transformed version (immutable) or self (mutable).
    """

    __slots__ = ()

    @abstractmethod
    def _clear(self):
        # type: (_BPC) -> _BPC
        """
        Clear.

        :return: Transformed.
        :rtype: objetto.bases.BaseProtectedCollection

        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()


# noinspection PyTypeChecker
_BIC = TypeVar("_BIC", bound="BaseInteractiveCollection")


# noinspection PyAbstractClass
class BaseInteractiveCollection(BaseProtectedCollection[T]):
    """
    Base interactive collection.

    Inherits from:
      - :class:`objetto.bases.BaseProtectedCollection`

    Inherited By:
      - :class:`objetto.bases.BaseInteractiveDict`
      - :class:`objetto.bases.BaseInteractiveList`
      - :class:`objetto.bases.BaseInteractiveSet`
      - :class:`objetto.bases.BaseState`
      - :class:`objetto.bases.BaseInteractiveStructure`

    Features:
      - Has public transformation methods.
      - Transformations return a transformed version (immutable) or self (mutable).
    """

    __slots__ = ()

    @final
    def clear(self):
        # type: (_BIC) -> _BIC
        """
        Clear.

        :return: Transformed.
        :rtype: objetto.bases.BaseInteractiveCollection
        """
        return self._clear()


# noinspection PyAbstractClass
class BaseMutableCollection(BaseProtectedCollection[T]):
    """
    Base mutable collection.

    Inherits from:
      - :class:`objetto.bases.BaseProtectedCollection`

    Inherited By:
      - :class:`objetto.bases.BaseMutableDict`
      - :class:`objetto.bases.BaseMutableList`
      - :class:`objetto.bases.BaseMutableSet`
      - :class:`objetto.bases.BaseMutableStructure`
      - :class:`objetto.bases.BaseProxyObject`

    Features:
      - Has public mutable transformation and magic methods.
    """

    __slots__ = ()

    def clear(self):
        # type: () -> None
        """Clear."""
        self._clear()



# -*- coding: utf-8 -*-
"""Object with state curated by attribute descriptors."""

from collections import Counter as ValueCounter
from contextlib import contextmanager
from inspect import getmro
from itertools import chain
from typing import TYPE_CHECKING, TypeVar, cast, overload
from weakref import WeakKeyDictionary

try:
    import collections.abc as collections_abc
except ImportError:
    import collections as collections_abc  # type: ignore

from six import iteritems, raise_from, with_metaclass

from .._applications import Application
from .._bases import FINAL_METHOD_TAG, MISSING, Base, final, init_context, make_base_cls
from .._changes import Update
from .._data import BaseData, Data, DataAttribute, InteractiveDictData
from .._states import DictState, SetState
from .._structures import (
    BaseAttribute,
    BaseAttributeMeta,
    BaseAttributeStructureMeta,
    BaseMutableAttributeStructure,
)
from ..utils.reraise_context import ReraiseContext
from ..utils.type_checking import (
    assert_is_callable,
    assert_is_instance,
    assert_is_subclass,
)
from ..utils.weak_reference import WeakReference
from .bases import (
    DELETED,
    BaseMutableObject,
    BaseObject,
    BaseObjectFunctions,
    BaseObjectMeta,
    Relationship,
)

if TYPE_CHECKING:
    from typing import (
        Any,
        Callable,
        Counter,
        Dict,
        Iterable,
        Iterator,
        List,
        Mapping,
        MutableMapping,
        Optional,
        Set,
        Tuple,
        Type,
        Union,
    )

    from .._applications import Store
    from .._history import HistoryObject
    from ..utils.factoring import LazyFactory

__all__ = ["Attribute", "Object"]


T = TypeVar("T")  # Any type.


# noinspection PyTypeChecker
_A = TypeVar("_A", bound="Attribute")


class AttributeMeta(BaseAttributeMeta):
    """Metaclass for `Attribute`."""

    @property
    @final
    def _relationship_type(cls):
        # type: () -> Type[Relationship]
        """Relationship type."""
        return Relationship


@final
class Attribute(with_metaclass(AttributeMeta, BaseAttribute[T])):
    """
    Attribute descriptor for :class:`objetto.objects.Object` classes.

    Inherits from:
      - :class:`objetto.bases.BaseAttribute`

    :param relationship: Relationship.
    :type relationship: objetto.objects.Relationship

    :param default: Default value.

    :param default_factory: Default value factory.
    :type default_factory: str or collections.abc.Callable or None

    :param module: Optional module path to use in case partial paths are provided.
    :type module: str or None

    :param required: Whether attribute is required to have a value or not.
    :type required: bool

    :param changeable: Whether attribute value can be changed.
    :type changeable: bool

    :param deletable: Whether attribute value can be deleted.
    :type deletable: bool

    :param finalized: If True, attribute can't be overridden by subclasses.
    :type finalized: bool

    :param abstracted: If True, attribute needs to be overridden by subclasses.
    :type abstracted: bool

    :param delegated: Whether attribute allows for delegates to be defined.
    :type delegated: bool

    :param dependencies: Attributes needed by the getter delegate.
    :type dependencies: collections.abc.Iterable[objetto.objects.Attribute] or \
objetto.objects.Attribute or None

    :param deserialize_to: Non-serialized attribute to deserialize this into.
    :type deserialize_to: objetto.objects.Attribute or None

    :raises TypeError: Invalid parameter type.
    :raises ValueError: Invalid parameter value.
    :raises ValueError: Can't declare same dependency more than once.
    :raises ValueError: Provided 'changeable' but 'delegated' is True.
    :raises ValueError: Provided 'deletable' but 'delegated' is True.
    :raises ValueError: Provided 'dependencies' but 'delegated' is False.
    :raises ValueError: Provided 'deserialize_to' but 'serialized' is False.
    :raises ValueError: Can't provide a serialized attribute to 'deserialize_to'.
    """

    __slots__ = (
        "__delegated",
        "__dependencies",
        "__deserialize_to",
        "__fget",
        "__fset",
        "__fdel",
        "__data_attribute",
    )

    def __init__(
        self,
        relationship=Relationship(),  # type: Relationship
        default=MISSING,  # type: Any
        default_factory=None,  # type: LazyFactory
        module=None,  # type: Optional[str]
        required=True,  # type: bool
        changeable=None,  # type: Optional[bool]
        deletable=None,  # type: Optional[bool]
        finalized=False,  # type: bool
        abstracted=False,  # type: bool
        delegated=False,  # type: bool
        dependencies=None,  # type: Optional[Union[Iterable[Attribute], Attribute]]
        deserialize_to=None,  # type: Optional[Attribute]
    ):
        # type: (...) -> None

        # 'changeable', 'deletable', 'delegated', and 'dependencies'
        if delegated:
            if dependencies is None:
                dependencies = ()
            else:
                with ReraiseContext(
                    (TypeError, ValueError), "'dependencies' parameter"
                ):
                    if isinstance(dependencies, collections_abc.Iterable):
                        visited_dependencies = set()
                        for dependency in dependencies:
                            assert_is_instance(dependency, Attribute, subtypes=False)
                            if dependency in visited_dependencies:
                                error = "can't declare same dependency more than once"
                                raise ValueError(error)
                            visited_dependencies.add(dependency)
                        dependencies = tuple(dependencies)
                    else:
                        assert_is_instance(dependencies, Attribute, subtypes=False)
                        dependencies = (dependencies,)
            if changeable is not None:
                error = "provided 'changeable' but 'delegated' is True"
                raise ValueError(error)
            else:
                changeable = False
            if deletable is not None:
                error = "provided 'deletable' but 'delegated' is True"
                raise ValueError(error)
            else:
                deletable = False
        else:
            if dependencies is not None:
                error = "provided 'dependencies' but 'delegated' is False"
                raise ValueError(error)
            else:
                dependencies = ()
            if changeable is None:
                changeable = True
            else:
                changeable = bool(changeable)
            if deletable is None:
                deletable = False
            else:
                deletable = bool(deletable)

        # 'deserialize_to'
        if deserialize_to is not None:
            if not relationship.serialized:
                error = "provided 'deserialize_to' but 'serialized' is False"
                raise ValueError(error)
            else:
                with ReraiseContext(TypeError, "'deserialize_to' parameter"):
                    assert_is_instance(deserialize_to, Attribute)
                if deserialize_to.relationship.serialized:
                    error = "can't provide a serialized attribute to 'deserialize_to'"
                    raise ValueError(error)

        super(Attribute, self).__init__(
            relationship=relationship,
            default=default,
            default_factory=default_factory,
            module=module,
            required=required,
            changeable=changeable,
            deletable=deletable,
            finalized=finalized,
            abstracted=abstracted,
        )

        self.__delegated = bool(delegated)
        self.__dependencies = dependencies
        self.__deserialize_to = deserialize_to
        self.__fget = None  # type: Optional[Callable]
        self.__fset = None  # type: Optional[Callable]
        self.__fdel = None  # type: Optional[Callable]
        self.__data_attribute = None  # type: Optional[DataAttribute]

    def __set__(self, instance, value):
        # type: (Object, T) -> None
        """
        Set attribute value.

        :param instance: Object instance.
        :type instance: objetto.objects.Object

        :param value: Value.
        """
        self.set_value(instance, value)

    def __delete__(self, instance):
        # type: (Object) -> None
        """
        Delete attribute value.

        :param instance: Object instance.
        :type instance: objetto.objects.Object
        """
        self.delete_value(instance)

    def to_dict(self):
        # type: () -> Dict[str, Any]
        """
        Convert to dictionary.

        :return: Dictionary.
        :rtype: dict[str, Any]
        """
        dct = super(Attribute, self).to_dict()
        dct.update(
            {
                "delegated": self.delegated,
                "dependencies": self.dependencies,
                "deserialize_to": self.deserialize_to,
                "fget": self.fget,
                "fset": self.fset,
                "fdel": self.fdel,
            }
        )
        return dct

    def set_value(self, instance, value):
        # type: (Object, T) -> None
        """
        Set attribute value.

        :param instance: Object instance.
        :type instance: objetto.objects.Object

        :param value: Value.
        """
        instance[self.get_name(instance)] = value

    def delete_value(self, instance):
        # type: (Object) -> None
        """
        Delete attribute value.

        :param instance: Object instance.
        :type instance: objetto.objects.Object
        """
        del instance[self.get_name(instance)]

    def getter(self, func):
        # type: (_A, Callable) -> _A
        """
        Define a getter delegate method.

        :param func: Delegate function.
        :type func: function

        :return: This attribute.
        :rtype: objetto.objects.Attribute

        :raises ValueError: Cannot define a getter for a non-delegated attribute.
        :raises ValueError: Getter delegate already defined.
        :raises TypeError: Invalid delegate type.
        """
        if not self.__delegated:
            error = "cannot define a getter for a non-delegated attribute"
            raise ValueError(error)
        if self.__fget is not None:
            error = "getter delegate already defined"
            raise ValueError(error)
        with ReraiseContext(TypeError, "attribute 'getter' delegate"):
            assert_is_callable(func)
        self.__fget = func
        return self

    def setter(self, func):
        # type: (_A, Callable) -> _A
        """
        Define a setter delegate method.

        :param func: Delegate function.
        :type func: function

        :return: This attribute.
        :rtype: objetto.objects.Attribute

        :raises ValueError: Cannot define a setter for a non-delegated attribute.
        :raises ValueError: Need to define a getter before defining a setter.
        :raises ValueError: Setter delegate already defined.
        :raises TypeError: Invalid delegate type.
        """
        if not self.__delegated:
            error = "cannot define a setter for a non-delegated attribute"
            raise ValueError(error)
        if self.__fget is None:
            error = "need to define a getter before defining a setter"
            raise ValueError(error)
        if self.__fset is not None:
            error = "setter delegate already defined"
            raise ValueError(error)
        with ReraiseContext(TypeError, "attribute 'setter' delegate"):
            assert_is_callable(func)
        self.__fset = func
        self._changeable = True
        return self

    def deleter(self, func):
        # type: (_A, Callable) -> _A
        """
        Define a deleter delegate method.

        :param func: Delegate function.
        :type func: function

        :return: This attribute.
        :rtype: objetto.objects.Attribute

        :raises ValueError: Cannot define a deleter for a non-delegated attribute.
        :raises ValueError: Need to define a getter before defining a deleter.
        :raises ValueError: Deleter delegate already defined.
        :raises TypeError: Invalid delegate type.
        """
        if not self.__delegated:
            error = "cannot define a deleter for a non-delegated attribute"
            raise ValueError(error)
        if self.__fget is None:
            error = "need to define a getter before defining a deleter"
            raise ValueError(error)
        if self.__fdel is not None:
            error = "deleter delegate already defined"
            raise ValueError(error)
        with ReraiseContext(TypeError, "attribute 'deleter' delegate"):
            assert_is_callable(func)
        self.__fdel = func
        self._deletable = True
        return self

    @property
    def relationship(self):
        # type: () -> Relationship
        """
        Relationship.

        :rtype: objetto.objects.Relationship
        """
        return cast("Relationship", super(Attribute, self).relationship)

    @property
    def delegated(self):
        # type: () -> bool
        """
        Whether attribute allows for delegates to be defined.

        :rtype: bool
        """
        return self.__delegated

    @property
    def dependencies(self):
        # type: () -> Tuple[Attribute, ...]
        """
        Attributes needed by the getter delegate.

        :rtype: tuple[objetto.objects.Attribute]
        """
        return self.__dependencies

    @property
    def deserialize_to(self):
        # type: () -> Optional[Attribute]
        """
        Non-serialized attribute to deserialize this into.

        :rtype: objetto.objects.Attribute or None
        """
        return self.__deserialize_to

    @property
    def fget(self):
        # type: () -> Optional[Callable]
        """
        Getter delegate.

        :rtype: function or None
        """
        return self.__fget

    @property
    def fset(self):
        # type: () -> Optional[Callable]
        """
        Setter delegate.

        :rtype: function or None
        """
        return self.__fset

    @property
    def fdel(self):
        # type: () -> Optional[Callable]
        """
        Deleter delegate.

        :rtype: function or None
        """
        return self.__fdel

    @property
    def data_attribute(self):
        # type: () -> Optional[DataAttribute]
        """
        Data attribute.

        :rtype: objetto.data.DataAttribute or None
        """
        if self.__data_attribute is None:
            data_relationship = self.relationship.data_relationship
            if data_relationship is not None:
                self.__data_attribute = DataAttribute(
                    data_relationship,
                    default=MISSING,
                    default_factory=None,
                    module=self.module,
                    required=False,
                    changeable=True,
                    deletable=True,
                    finalized=False,
                    abstracted=False,
                )
        return self.__data_attribute


@final
class Functions(BaseObjectFunctions):
    """Static functions for `Object`."""

    __slots__ = ()

    @staticmethod
    def replace_child_data(store, child, data_location, new_child_data):
        # type: (Store, BaseObject, Any, BaseData) -> Store
        """
        Replace child data.

        :param store: Object's store.
        :param child: Child getting their data replaced.
        :param data_location: Location of the existing child's data.
        :param new_child_data: New child's data.
        :return: Updated object's store.
        """
        original_data = store.data
        assert original_data is not None
        data = cast("Data", original_data)._set(data_location, new_child_data)
        return store.set("data", data)

    @staticmethod
    def get_initial(
        obj,  # type: Object
        input_values,  # type: Mapping[str, Any]
        factory=True,  # type: bool
    ):
        # type: (...) -> Mapping[str, Any]
        """
        Get initial values.

        :param obj: Object.
        :param input_values: Input values.
        :param factory: Whether to run values through factory.
        :return: Initial values.
        """
        cls = type(obj)
        initial = {}
        kwargs = {"app": obj.app}

        for name, value in iteritems(input_values):
            try:
                attribute = cls._attributes[name]
            except KeyError:
                error = "'{}' has no attribute '{}'".format(cls.__fullname__, name)
                exc = AttributeError(error)
                raise_from(exc, None)
                raise exc
            initial[name] = attribute.relationship.fabricate_value(
                value, factory=factory, **kwargs
            )

        for name, attribute in iteritems(cls._attributes):
            if name not in initial:
                if attribute.has_default:
                    initial[name] = attribute.fabricate_default_value(**kwargs)

        return initial

    @staticmethod
    def check_missing(cls, state):
        # type: (Type[Object], DictState[str, Any]) -> None
        """
        Check for attributes with no values.

        :param cls: Object class.
        :param state: State.
        :raises TypeError: Raised when required attributes are missing.
        """
        missing_attributes = set(cls._attributes).difference(state)  # type: Set[str]
        optional_attributes = set(
            n for n, a in iteritems(cls._attributes) if not a.required
        )  # type: Set[str]
        if missing_attributes.difference(optional_attributes):
            error = "missing required attribute{} {}".format(
                "s" if len(missing_attributes) != 1 else "",
                ", ".join("'{}'".format(n) for n in missing_attributes),
            )
            raise TypeError(error)

    @staticmethod
    def update(obj, input_values, factory=True):
        # type: (Object, Mapping[str, Any], bool) -> None
        """
        Update object with values.

        :param obj: Object.
        :param input_values: Input values.
        :param factory: Whether to run values through factory.
        """
        cls = type(obj)
        with obj.app.__.write_context(obj) as (read, write):
            if not input_values:
                return

            # Set updates through intermediary object, sort by number of dependencies.
            intermediary_object = IntermediaryObject(obj.app, type(obj), read().state)
            sorted_input_values = sorted(
                iteritems(input_values),
                key=lambda i: len(cls._attribute_flattened_dependencies.get(i[0], ())),
            )
            for name, value in sorted_input_values:
                try:
                    attribute = cls._attributes[name]
                except KeyError:
                    error = "'{}' has no attribute '{}'".format(cls.__fullname__, name)
                    exc = AttributeError(error)
                    raise_from(exc, None)
                    raise exc
                else:
                    if value is DELETED:
                        intermediary_object.__.delete_value(name)
                    else:
                        if factory:
                            value = attribute.relationship.fabricate_value(
                                value, factory=True, **{"app": obj.app}
                            )
                    intermediary_object.__.set_value(name, value, factory=False)

            # Get results from changes.
            new_values, old_values = intermediary_object.__.get_results()

            # Process raw updates.
            Functions.raw_update(obj, new_values, old_values)

    @staticmethod
    def raw_update(
        obj,  # type: Object
        new_values,  # type: Mapping[str, Any]
        old_values,  # type: Mapping[str, Any]
        history=None,  # type: Optional[HistoryObject]
    ):
        # type: (...) -> None
        """
        Raw update (without going through intermediary object).

        :param obj: Object.
        :param new_values: New values.
        :param old_values: Old values.
        :param history: History than triggered this change during undo/redo.
        :raises AttributeError: Attribute is not changeable and already has a value.
        :raises AttributeError: Attribute is not deletable.
        """
        cls = type(obj)
        with obj.app.__.write_context(obj) as (read, write):

            # Get state, data, and locations cache.
            store = read()
            state = old_state = store.state  # type: DictState[str, Any]
            data = store.data  # type: Data
            metadata = store.metadata  # type: InteractiveDictData[str, Any]
            locations = metadata.get(
                "locations", DictState()
            )  # type: DictState[BaseObject, str]

            # Prepare change information.
            child_counter = ValueCounter()  # type: Counter[BaseObject]
            old_children = set()  # type: Set[BaseObject]
            new_children = set()  # type: Set[BaseObject]
            history_adopters = set()  # type: Set[BaseObject]

            # For every new value.
            for name, value in iteritems(new_values):

                # Get attribute and relationship.
                attribute = cls._attributes[name]
                relationship = attribute.relationship

                # Are we deleting it?
                delete_item = value is DELETED
                if delete_item and not attribute.deletable and not attribute.delegated:
                    error = "attribute '{}' is not deletable".format(name)
                    raise AttributeError(error)

                # Get old value.
                old_value = old_values[name]

                # Child relationship.
                if relationship.child:
                    same_app = not delete_item and obj._in_same_application(value)

                    # Update children counter, old/new children sets, and locations.
                    if old_value is not DELETED:
                        if not attribute.changeable and not attribute.delegated:
                            error = (
                                "non-changeable attribute '{}' already has a value"
                            ).format(name)
                            raise AttributeError(error)
                        if obj._in_same_application(old_value):
                            child_counter[old_value] -= 1
                            old_children.add(old_value)
                            locations = locations.remove(old_value)
                    if same_app:
                        child_counter[value] += 1
                        new_children.add(value)
                        locations = locations.set(value, name)

                    # Add history adopter.
                    if relationship.history and same_app:
                        history_adopters.add(value)

                    # Update data.
                    if relationship.data:
                        if delete_item:
                            data = data._delete(name)
                        else:
                            data_relationship = relationship.data_relationship
                            assert data_relationship is not None
                            if same_app:
                                with value.app.__.write_context(value) as (v_read, _):
                                    data = data._set(
                                        name,
                                        data_relationship.fabricate_value(
                                            v_read().data
                                        ),
                                    )
                            else:
                                data = data._set(
                                    name,
                                    data_relationship.fabricate_value(value),
                                )

                # Update state.
                if not delete_item:
                    state = state.set(name, value)
                else:
                    state = state.remove(name)

            # Store locations in the metadata.
            metadata = metadata.set("locations", locations)

            # Prepare change.
            change = Update(
                __redo__=Functions.redo_raw_update,
                __undo__=Functions.undo_raw_update,
                obj=obj,
                old_children=old_children,
                new_children=new_children,
                old_values=old_values,
                new_values=new_values,
                old_state=old_state,
                new_state=state,
                history_adopters=history_adopters,
                history=history,
            )
            write(state, data, metadata, child_counter, change)

    @staticmethod
    def redo_raw_update(change):
        # type: (Update) -> None
        """
        Raw update object state (REDO).

        :param change: Change.
        """
        Functions.raw_update(
            cast("Object", change.obj),
            change.new_values,
            change.old_values,
            history=change.obj._history,
        )

    @staticmethod
    def undo_raw_update(change):
        # type: (Update) -> None
        """
        Raw update object state (UNDO).

        :param change: Change.
        """
        Functions.raw_update(
            cast("Object", change.obj),
            change.old_values,
            change.new_values,
            history=change.obj._history,
        )


# Mark 'Functions' as a final member.
type.__setattr__(cast(type, Functions), FINAL_METHOD_TAG, True)


class ObjectMeta(BaseAttributeStructureMeta, BaseObjectMeta):
    """
    Metaclass for `Object`.

    :raises TypeError: Attribute is delegated but no delegates were defined.
    :raises TypeError: Attribute declares a dependency which is not available.
    """

    __data_type = WeakKeyDictionary({})  # type: MutableMapping[ObjectMeta, Type[Data]]
    __attribute_dependencies = WeakKeyDictionary(
        {}
    )  # type: MutableMapping[ObjectMeta, DictState]
    __attribute_dependents = WeakKeyDictionary(
        {}
    )  # type: MutableMapping[ObjectMeta, DictState]
    __attribute_flattened_dependencies = WeakKeyDictionary(
        {}
    )  # type: MutableMapping[ObjectMeta, DictState]
    __attribute_flattened_dependents = WeakKeyDictionary(
        {}
    )  # type: MutableMapping[ObjectMeta, DictState]

    def __init__(cls, name, bases, dct):
        # type: (str, Tuple[Type, ...], Dict[str, Any]) -> None
        super(ObjectMeta, cls).__init__(name, bases, dct)

        # Check and gather attribute dependencies.
        dependencies = dict(
            (n, SetState()) for n in cls._attributes
        )  # type: Dict[str, SetState[str]]
        dependents = dict(
            (n, SetState()) for n in cls._attributes
        )  # type: Dict[str, SetState[str]]
        for attribute_name, attribute in iteritems(cls._attributes):

            # Delegated, but does not define any delegate.
            if attribute.delegated and attribute.fget is None:
                error = (
                    "attribute '{}.{}' is delegated but no delegates were defined"
                ).format(name, attribute_name)
                raise TypeError(error)

            # Dependencies.
            for dependency in attribute.dependencies:

                # Dependency is not in the same object.
                if dependency not in cls._attribute_names:
                    error = (
                        "attribute '{}.{}' declares an attribute as a dependency which "
                        "is not a valid attribute in the same object"
                    ).format(name, attribute_name)
                    raise TypeError(error)

                # Store dependencies.
                dependency_name = cls._attribute_names[dependency]
                dependencies[attribute_name] = dependencies.get(
                    attribute_name, SetState()
                ).add(dependency_name)
                dependents[dependency_name] = dependents.get(
                    dependency_name, SetState()
                ).add(attribute_name)

        def _resolve(nm, deps):
            return deps[nm].update(
                chain.from_iterable(_resolve(n, deps) for n in deps[nm])
            )

        def _flatten(deps):
            flattened = {}
            for nm in deps:
                flattened[nm] = _resolve(nm, deps)
            return flattened

        # Store dependencies and dependents.
        type(cls).__attribute_dependencies[cls] = DictState(dependencies)
        type(cls).__attribute_dependents[cls] = DictState(dependents)
        type(cls).__attribute_flattened_dependencies[cls] = DictState(
            _flatten(dependencies)
        )
        type(cls).__attribute_flattened_dependents[cls] = DictState(
            _flatten(dependents)
        )

    @property
    @final
    def _attribute_type(cls):
        # type: () -> Type[Attribute]
        """Attribute type."""
        return Attribute

    @property
    @final
    def _attributes(cls):
        # type: () -> Mapping[str, Attribute]
        """Attributes mapped by name."""
        return cast("Mapping[str, Attribute]", super(ObjectMeta, cls)._attributes)

    @property
    @final
    def _attribute_names(cls):
        # type: () -> Mapping[Attribute, str]
        """Names mapped by attribute."""
        return cast("Mapping[Attribute, str]", super(ObjectMeta, cls)._attribute_names)

    @property
    @final
    def _state_factory(cls):
        # type: () -> Callable[..., DictState]
        """State factory."""
        return DictState

    @property
    @final
    def _attribute_dependencies(cls):
        # type: () -> DictState[str, SetState[str]]
        """Attribute dependencies."""
        return type(cls).__attribute_dependencies[cls]

    @property
    @final
    def _attribute_dependents(cls):
        # type: () -> DictState[str, SetState[str]]
        """Attribute dependents."""
        return type(cls).__attribute_dependents[cls]

    @property
    @final
    def _attribute_flattened_dependencies(cls):
        # type: () -> DictState[str, SetState[str]]
        """Flattened attribute dependencies."""
        return type(cls).__attribute_flattened_dependencies[cls]

    @property
    @final
    def _attribute_flattened_dependents(cls):
        # type: () -> DictState[str, SetState[str]]
        """Flattened attribute dependents."""
        return type(cls).__attribute_flattened_dependents[cls]

    @property
    @final
    def Data(cls):
        # type: () -> Type[Data]
        """Data type."""

        # Try to get cached data type.
        mcs = type(cls)
        try:
            data_type = mcs.__data_type[cls]
        except KeyError:
            user_data_type = None
            user_data_type_owner = None
            for base in reversed(getmro(cls)):
                if "Data" in base.__dict__:
                    user_data_type = base.__dict__["Data"]
                    user_data_type_owner = base

            # User-defined data type.
            if user_data_type is not None:
                assert user_data_type_owner is not None
                with ReraiseContext(
                    TypeError,
                    "custom 'Data' class member defined in '{}'".format(
                        user_data_type_owner.__name__
                    ),
                ):
                    assert_is_subclass(user_data_type, Data)
                mcs.__data_type[cls] = data_type = user_data_type

            # Automatically defined data type.
            else:

                # Build data attributes.
                attributes = {}
                for attribute_name, attribute in iteritems(cls._attributes):
                    if attribute.relationship.data:
                        data_attribute = attribute.data_attribute
                        if data_attribute is None:
                            continue
                        attributes[attribute_name] = data_attribute

                # Prepare dct.
                dct = {}  # type: Dict[str, Any]
                dct.update(cls._data_methods)
                dct.update(attributes)
                if cls._unique_descriptor is not None:
                    assert cls._unique_descriptor_name is not None
                    dct[cls._unique_descriptor_name] = cls._unique_descriptor

                # Build data type and cache it.
                data_type = make_base_cls(
                    base=Data,
                    qual_name="{}.{}".format(cls.__fullname__, "Data"),
                    module=cls.__module__,
                    dct=dct,
                )
                mcs.__data_type[cls] = data_type

        return data_type


# noinspection PyTypeChecker
_O = TypeVar("_O", bound="Object")


class Object(
    with_metaclass(ObjectMeta, BaseMutableObject[str], BaseMutableAttributeStructure)
):
    """
    Object.

    Inherits from:
      - :class:`objetto.bases.BaseMutableObject`
      - :class:`objetto.bases.BaseMutableAttributeStructure`

    Inherited by:
      - :class:`objetto.history.HistoryObject`
      - :class:`objetto.history.BatchChanges`

    :param app: Application.
    :type app: objetto.applications.Application

    :param initial: Initial values.
    """

    __slots__ = ()
    __functions__ = Functions

    def __init__(self, app, **initial):
        # type: (Application, Any) -> None
        super(Object, self).__init__(app=app)
        cls = type(self)
        with self.app.write_context():
            self.__functions__.update(
                self,
                self.__functions__.get_initial(self, initial),
                factory=False,
            )
            self.__functions__.check_missing(cls, self._state)

    @classmethod
    @final
    def _get_relationship(cls, location):
        # type: (str) -> Relationship
        """
        Get relationship at location (attribute name).

        :param location: Location (attribute name).
        :type location: str

        :return: Relationship.
        :rtype: objetto.objects.Relationship

        :raises KeyError: Attribute does not exist.
        """
        return cast("Relationship", cls._get_attribute(location).relationship)

    @classmethod
    @final
    def _get_attribute(cls, name):
        # type: (str) -> Attribute
        """
        Get attribute by name.

        :param name: Attribute name.
        :type name: str

        :return: Attribute.
        :rtype: objetto.objects.Attribute

        :raises KeyError: Attribute does not exist.
        """
        return cast("Attribute", cls._attributes[name])

    @final
    def _clear(self):
        # type: (_O) -> _O
        """
        Clear deletable attribute values.

        :return: Transformed.
        :rtype: objetto.objects.Object

        :raises AttributeError: No deletable attributes.
        """
        with self.app.write_context():
            cls = type(self)
            update = {}
            for name in self._state:
                attribute = cls._get_attribute(name)
                if attribute.deletable:
                    update[name] = DELETED
            if not update:
                error = "'{}' has no deletable attributes".format(
                    type(self).__fullname__
                )
                raise AttributeError(error)
            self._update(update)
        return self

    @overload
    def _update(self, __m, **kwargs):
        # type: (_O, Mapping[str, Any], Any) -> _O
        pass

    @overload
    def _update(self, __m, **kwargs):
        # type: (_O, Iterable[Tuple[str, Any]], Any) -> _O
        pass

    @overload
    def _update(self, **kwargs):
        # type: (_O, Any) -> _O
        pass

    @final
    def _update(self, *args, **kwargs):
        """
        Update multiple attribute values.
        Same parameters as :meth:`dict.update`.

        :return: Transformed.
        :rtype: objetto.objects.Object

        :raises AttributeError: Attribute is not changeable and already has a value.
        """
        self.__functions__.update(self, dict(*args, **kwargs))
        return self

    @final
    def _set(self, name, value):
        # type: (_O, str, Any) -> _O
        """
        Set attribute value.

        :param name: Attribute name.
        :type name: str

        :param value: Value.

        :return: Transformed.
        :rtype: objetto.objects.Object

        :raises AttributeError: Attribute is not changeable and already has a value.
        """
        self.__functions__.update(self, {name: value})
        return self

    @final
    def _delete(self, name):
        # type: (_O, str) -> _O
        """
        Delete attribute value.

        :param name: Attribute name.
        :type name: str

        :return: Transformed.
        :rtype: objetto.objects.Object

        :raises KeyError: Attribute does not exist or has no value.
        :raises AttributeError: Attribute is not deletable.
        """
        self.__functions__.update(self, {name: DELETED})
        return self

    @final
    def _locate(self, child):
        # type: (BaseObject) -> str
        """
        Locate child object.

        :param child: Child object.
        :type child: objetto.bases.BaseObject

        :return: Location.
        :rtype: str

        :raises ValueError: Could not locate child.
        """
        with self.app.__.read_context(self) as read:
            metadata = read().metadata
            try:
                return metadata["locations"][child]
            except KeyError:
                error = "could not locate child {} in {}".format(child, self)
                exc = ValueError(error)
                raise_from(exc, None)
                raise exc

    @final
    def _locate_data(self, child):
        # type: (BaseObject) -> str
        """
        Locate child object's data.

        :param child: Child object.
        :type child: objetto.bases.BaseObject

        :return: Data location.
        :rtype: str

        :raises ValueError: Could not locate child's data.
        """
        return self._locate(child)

    @classmethod
    @final
    def deserialize(cls, serialized, app=None, **kwargs):
        # type: (Type[_O], Dict[str, Any], Application, Any) -> _O
        """
        Deserialize.

        :param serialized: Serialized.
        :type serialized: dict[str, Any]

        :param app: Application (required).
        :type app: objetto.applications.Application

        :param kwargs: Keyword arguments to be passed to the deserializers.

        :return: Deserialized.
        :rtype: objetto.objects.Object

        :raises ValueError: Missing required 'app' attribute.
        """
        if app is None:
            error = (
                "missing required 'app' keyword argument for '{}.deserialize()' method"
            ).format(cls.__fullname__)
            raise ValueError(error)
        kwargs["app"] = app

        with app.write_context():
            self = cast("_O", cls.__new__(cls))
            with init_context(self):
                super(Object, self).__init__(app)

                initial = {}  # type: Dict[str, Any]
                for name, value in iteritems(serialized):
                    try:
                        attribute = cls._attributes[name]
                    except KeyError:
                        error = "'{}.deserialize'; '{}' has no attribute '{}'".format(
                            cls.__fullname__,
                            cls.__fullname__,
                            name,
                        )
                        exc = AttributeError(error)
                        raise_from(exc, None)
                        raise exc

                    if attribute.deserialize_to is not None:
                        deserialize_to_name = cls._attribute_names[
                            attribute.deserialize_to
                        ]
                    else:
                        deserialize_to_name = name

                    initial[deserialize_to_name] = cls.deserialize_value(
                        value, name, **kwargs
                    )

                self.__functions__.update(
                    self,
                    self.__functions__.get_initial(self, initial),
                    factory=False,
                )
                Functions.check_missing(cls, self._state)
            return self

    @final
    def serialize(self, **kwargs):
        # type: (Any) -> Dict[str, Any]
        """
        Serialize.

        :param kwargs: Keyword arguments to be passed to the serializers.

        :return: Serialized.
        :rtype: dict[str, Any]
        """
        with self.app.read_context():
            return dict(
                (k, self.serialize_value(v, k, **kwargs))
                for k, v in iteritems(self._state)
                if type(self)._get_relationship(k).serialized
            )

    @property
    @final
    def _state(self):
        # type: () -> DictState[str, Any]
        """
        State.

        :rtype: objetto.states.DictState[str, Any]
        """
        return cast("DictState", super(Object, self)._state)

    @property
    @final
    def data(self):
        # type: () -> Data
        """
        Data.

        :rtype: objetto.data.Data
        """
        return cast("Data", super(Object, self).data)


@final
class IntermediaryObjectInternals(Base):
    """
    Internals for `IntermediaryObject`.

    :param iobj: Internal object.
    :param app: Application.
    :param cls: Object class.
    :param state: Object state.
    """

    __slots__ = (
        "__iobj_ref",
        "__app",
        "__cls",
        "__state",
        "__dependencies",
        "__in_getter",
        "__new_values",
        "__old_values",
        "__dirty",
    )

    def __init__(
        self,
        iobj,  # type: IntermediaryObject
        app,  # type: Application
        cls,  # type: Type[Object]
        state,  # type: DictState[str, Any]
    ):
        # type: (...) -> None
        self.__iobj_ref = WeakReference(iobj)
        self.__app = app
        self.__cls = cls
        self.__state = state
        self.__dependencies = None  # type: Optional[Tuple[Attribute, ...]]
        self.__in_getter = None  # type: Optional[Attribute]
        self.__new_values = {}  # type: Dict[str, Any]
        self.__old_values = {}  # type: Dict[str, Any]
        self.__dirty = set(cls._attributes).difference(state)  # type: Set[str]

    def get_value(self, name):
        """
        Get current value for attribute.

        :param name: Attribute name.
        :return: Value.
        :raises NameError: Can't access attribute not declared as dependency.
        :raises AttributeError: Attribute has no value.
        """
        attribute = self.__get_attribute(name)
        if self.__dependencies is not None and attribute not in self.__dependencies:
            error = (
                "can't access '{}' attribute from '{}' getter delegate since it was "
                "not declared as a dependency"
            ).format(name, self.__cls._attribute_names[self.__in_getter])
            raise NameError(error)

        if name in self.__dirty:
            value = MISSING
        else:
            try:
                value = self.__new_values[name]
            except KeyError:
                try:
                    value = self.__state[name]
                except KeyError:
                    value = MISSING

        if value in (MISSING, DELETED):
            if attribute.delegated:
                with self.__getter_context(attribute):
                    value = attribute.fget(self.iobj)
                value = attribute.relationship.fabricate_value(
                    value, factory=True, **{"app": self.app}
                )
                self.__set_new_value(name, value)
                return value
            else:
                error = "attribute '{}' has no value".format(name)
                raise AttributeError(error)
        else:
            return value

    def set_value(self, name, value, factory=True):
        # type: (str, Any, bool) -> None
        """
        Set attribute value.

        :param name: Attribute name.
        :param value: Value.
        :param factory: Whether to run value through factory.
        :raises AttributeError: Can't set attributes while running getter delegate.
        :raises AttributeError: Attribute is read-only.
        :raises AttributeError: Attribute already has a value and can't be changed.
        :raises AttributeError: Can't delete attributes while running getter delegate.
        :raises AttributeError: Attribute is not deletable.
        """

        if value is DELETED:
            self.delete_value(name)
            return

        if self.__in_getter is not None:
            error = "can't set attributes while running getter delegate"
            raise AttributeError(error)

        attribute = self.__get_attribute(name)
        if not attribute.changeable:
            if attribute.delegated:
                error = "attribute '{}' is read-only".format(name)
                raise AttributeError(error)
            try:
                self.get_value(name)
            except AttributeError:
                pass
            else:
                error = (
                    "attribute '{}' already has a value and can't be changed"
                ).format(name)
                raise AttributeError(error)

        if factory:
            value = attribute.relationship.fabricate_value(
                value, factory=True, **{"app": self.app}
            )
        if attribute.delegated:
            attribute.fset(self.iobj, value)
        else:
            self.__set_new_value(name, value)

    def delete_value(self, name):
        """
        Delete attribute.

        :param name: Attribute name.
        :raises AttributeError: Can't delete attributes while running getter delegate.
        :raises AttributeError: Attribute is not deletable.
        """

        if self.__in_getter is not None:
            error = "can't delete attributes while running getter delegate"
            raise AttributeError(error)

        attribute = self.__get_attribute(name)
        if not attribute.deletable:
            error = "attribute '{}' is not deletable".format(name)
            raise AttributeError(error)

        if attribute.delegated:
            attribute.fdel(self.iobj)
        else:
            self.get_value(name)  # will error if has no value, which we want
            self.__set_new_value(name, DELETED)

    @contextmanager
    def __getter_context(self, attribute):
        # type: (Attribute) -> Iterator
        """
        Getter context.

        :param attribute: Attribute.
        :return: Getter context manager.
        """
        before = self.__in_getter
        before_dependencies = self.__dependencies

        self.__in_getter = attribute
        if attribute.delegated:
            self.__dependencies = attribute.dependencies
        else:
            self.__dependencies = None

        try:
            yield
        finally:
            self.__in_getter = before
            self.__dependencies = before_dependencies

    def __set_new_value(self, name, value):
        # type: (str, Any) -> None
        """
        Set new attribute value.

        :param name: Attribute name.
        :param value: Value.
        """
        try:
            old_value = self.__state[name]
        except KeyError:
            old_value = DELETED

        if value is not old_value:
            self.__old_values[name] = old_value
            self.__new_values[name] = value
        else:
            self.__old_values.pop(name, None)
            self.__new_values.pop(name, None)

        self.__dirty.discard(name)
        for dependent in self.__cls._attribute_flattened_dependents[name]:
            self.__dirty.add(dependent)
            try:
                old_value = self.__state[dependent]
            except KeyError:
                self.__new_values.pop(dependent, None)
                self.__old_values.pop(dependent, None)
            else:
                self.__old_values[dependent] = old_value
                self.__new_values[dependent] = DELETED

    def __get_attribute(self, name):
        # type: (str) -> Any
        """
        Get attribute value.

        :param name: Attribute name.
        :return: Value.
        :raises AttributeError: Has no such attribute.
        """
        try:
            return self.cls._attributes[name]
        except KeyError:
            pass
        error = "'{}' has no attribute '{}'".format(self.cls.__fullname__, name)
        raise AttributeError(error)

    def get_results(self):
        # type: () -> Tuple[Mapping[str, Any], Mapping[str, Any]]
        """
        Get results.

        :return: New values, old values.
        """
        sorted_dirty = sorted(
            self.__dirty,
            key=lambda n: len(self.__cls._attribute_flattened_dependencies[n]),
        )
        failed = set()
        success = set()
        for name in sorted_dirty:
            try:
                self.get_value(name)
            except AttributeError:
                failed.add(name)
            else:
                success.add(name)

        new_values = self.__new_values.copy()
        old_values = self.__old_values.copy()

        return new_values, old_values

    @property
    def iobj(self):
        # type: () -> Optional[IntermediaryObject]
        """Intermediary object."""
        return self.__iobj_ref()

    @property
    def app(self):
        # type: () -> Application
        """Application."""
        return self.__app

    @property
    def cls(self):
        # type: () -> Type[Object]
        """Object class."""
        return self.__cls

    @property
    def state(self):
        # type: () -> DictState[str, Any]
        """Object state."""
        return self.__state

    @property
    def in_getter(self):
        # type: () -> Optional[Attribute]
        """Whether running in an attribute's getter delegate."""
        return self.__in_getter


@final
class IntermediaryObject(Base):
    """
    Intermediary object provided to delegates.

    :param app: Application.
    :param cls: Object class.
    :param state: Object state.
    """

    __slots__ = ("__weakref__", "__")

    def __init__(self, app, cls, state):
        # type: (Application, Type[Object], DictState[str, Any]) -> None
        object.__setattr__(
            self,
            "__",
            IntermediaryObjectInternals(self, app, cls, state),
        )

    def __dir__(self):
        # type: () -> List[str]
        """
        Get attribute names.

        :return: Attribute names.
        """
        if self.__.in_getter is not None:
            attribute = self.__.in_getter
            return sorted(
                n
                for n, a in iteritems(self.__.cls._attributes)
                if a is attribute or a in a.dependencies
            )
        return sorted(self.__.cls._attributes)

    def __getattr__(self, name):
        # type: (str) -> Any
        """
        Get attribute value.

        :param name: Attribute name.
        :return: Value.
        """
        if name != "__" and name in self.__.cls._attributes:
            return self[name]
        else:
            return self.__getattribute__(name)

    def __setattr__(self, name, value):
        # type: (str, Any) -> None
        """
        Set attribute value.

        :param name: Attribute name.
        :param value: Value.
        """
        if name in self.__.cls._attributes:
            self[name] = value
        else:
            super(IntermediaryObject, self).__setattr__(name, value)

    def __delattr__(self, name):
        # type: (str) -> None
        """
        Delete attribute value.

        :param name: Attribute name.
        """
        if name in self.__.cls._attributes:
            del self[name]
        else:
            super(IntermediaryObject, self).__delattr__(name)

    def __getitem__(self, name):
        # type: (str) -> Any
        """
        Get attribute value.

        :param name: Attribute name.
        :return: Value.
        """
        return self.__.get_value(name)

    def __setitem__(self, name, value):
        # type: (str, Any) -> None
        """
        Set attribute value.

        :param name: Attribute name.
        :param value: Value.
        """
        self.__.set_value(name, value)

    def __delitem__(self, name):
        # type: (str) -> None
        """
        Delete attribute value.

        :param name: Attribute name.
        """
        self.__.delete_value(name)



# -*- coding: utf-8 -*-
"""Factories."""

from ._factories import (
    Boolean,
    Curated,
    FloatingPoint,
    Integer,
    MultiFactory,
    RegexMatch,
    RegexSub,
    String,
)

__all__ = [
    "MultiFactory",
    "Integer",
    "FloatingPoint",
    "RegexMatch",
    "RegexSub",
    "String",
    "Curated",
    "Boolean",
]



# -*- coding: utf-8 -*-
"""Immutable list state."""

from itertools import chain
from typing import TYPE_CHECKING, TypeVar, cast, overload

try:
    import collections.abc as collections_abc
except ImportError:
    import collections as collections_abc  # type: ignore

from pyrsistent import pvector
from six import iteritems

from .._bases import BaseInteractiveList, final
from ..utils.custom_repr import custom_iterable_repr, custom_mapping_repr
from ..utils.list_operations import pre_move, resolve_continuous_slice, resolve_index
from ..utils.recursive_repr import recursive_repr
from .bases import BaseState

if TYPE_CHECKING:
    from typing import Any, Iterable, Iterator, Optional, Tuple, Type, Union

    from pyrsistent.typing import PVector

__all__ = ["ListState"]


T = TypeVar("T")  # Any type.


# noinspection PyTypeChecker
_LS = TypeVar("_LS", bound="ListState")


@final
class ListState(BaseState[T], BaseInteractiveList[T]):
    """
    Immutable list state.

    Inherits from:
      - :class:`objetto.bases.BaseState`
      - :class:`objetto.bases.BaseInteractiveList`

    :param initial: Initial values.
    :type initial: collections.abc.Iterable
    """

    __slots__ = ()

    @classmethod
    def _make(cls, internal=pvector()):
        # type: (Type[_LS], PVector[T]) -> _LS
        """
        Make new state by directly setting the internal state.

        :param internal: Internal state.
        :return: State.
        """
        return super(ListState, cls)._make(internal)

    @staticmethod
    def _make_internal(initial):
        # type: (Iterable[T]) -> PVector[T]
        """
        Initialize internal state.

        :param initial: Initial values.
        """
        return pvector(initial)

    def __init__(self, initial=()):
        # type: (Iterable[T]) -> None
        super(ListState, self).__init__(initial=initial)

    def __hash__(self):
        # type: () -> int
        """
        Get hash.

        :return: Hash.
        :rtype: int
        """
        return super(ListState, self).__hash__()

    def __eq__(self, other):
        # type: (object) -> bool
        """
        Compare for equality.

        :param other: Another object.

        :return: True if equal.
        :rtype: bool
        """
        if self is other:
            return True
        if not isinstance(other, collections_abc.Hashable):
            return self._internal == other
        if isinstance(other, ListState):
            return self._internal == other._internal
        return False

    def __contains__(self, value):
        # type: (Any) -> bool
        """
        Get whether value is present.

        :param value: Value.

        :return: True if contains.
        :rtype: bool
        """
        return value in self._internal

    def __iter__(self):
        # type: () -> Iterator[T]
        """
        Iterate over values.

        :return: Values iterator.
        :rtype: collections.abc.Iterator
        """
        for value in self._internal:
            yield value

    def __len__(self):
        # type: () -> int
        """
        Get value count.

        :return: Value count.
        :rtype: int
        """
        return len(self._internal)

    @recursive_repr
    def __repr__(self):
        # type: () -> str
        """
        Get representation.

        :return: Representation.
        :rtype: str
        """
        return custom_iterable_repr(
            self._internal,
            prefix="{}([".format(type(self).__fullname__),
            suffix="])",
        )

    def __reversed__(self):
        # type: () -> Iterator[T]
        """
        Iterate over reversed values.

        :return: Reversed values iterator.
        :rtype: collections.abc.Iterator
        """
        return reversed(self._internal)

    @overload
    def __getitem__(self, index):
        # type: (int) -> T
        pass

    @overload
    def __getitem__(self, index):
        # type: (slice) -> ListState[T]
        pass

    def __getitem__(self, index):
        """
        Get value/values at index/from slice.

        :param index: Index/slice.
        :type index: int or slice

        :return: Value/values.
        :rtype: Any or objetto.states.ListState
        """
        if isinstance(index, slice):
            return self._make(self._internal[index])
        else:
            return self._internal[index]

    def _clear(self):
        # type: (_LS) -> _LS
        """
        Clear.

        :return: Transformed.
        :rtype: objetto.states.ListState
        """
        return self._make()

    def _insert(self, index, *values):
        # type: (_LS, int, T) -> _LS
        """
        Insert value(s) at index.

        :param index: Index.
        :type index: int

        :param values: Value(s).

        :return: Transformed.
        :rtype: objetto.states.ListState

        :raises ValueError: No values provided.
        """
        if not values:
            error = "no values provided"
            raise ValueError(error)
        index = self.resolve_index(index, clamp=True)
        if index == len(self._internal):
            return self.extend(values)
        elif index == 0:
            return self._make(pvector(values) + self._internal)
        else:
            return self._make(
                self._internal[:index] + pvector(values) + self._internal[index:]
            )

    def _append(self, value):
        # type: (_LS, T) -> _LS
        """
        Append value at the end.

        :param value: Value.

        :return: Transformed.
        :rtype: objetto.states.ListState
        """
        return self._make(self._internal.append(value))

    def _extend(self, iterable):
        # type: (_LS, Iterable[T]) -> _LS
        """
        Extend at the end with iterable.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :return: Transformed.
        :rtype: objetto.states.ListState
        """
        return self._make(self._internal.extend(iterable))

    def _remove(self, value):
        # type: (_LS, T) -> _LS
        """
        Remove first occurrence of value.

        :param value: Value.

        :return: Transformed.
        :rtype: objetto.states.ListState

        :raises ValueError: Value is not present.
        """
        return self._make(self._internal.remove(value))

    def _reverse(self):
        # type: (_LS) -> _LS
        """
        Reverse values.

        :return: Transformed.
        :rtype: objetto.states.ListState
        """
        return self._make(pvector(reversed(self._internal)))

    def _move(self, item, target_index):
        # type: (_LS, Union[slice, int], int) -> _LS
        """
        Move values internally.

        :param item: Index/slice.
        :type item: int or slice

        :param target_index: Target index.
        :type target_index: int

        :return: Transformed.
        :rtype: objetto.states.ListState
        """
        result = pre_move(len(self._internal), item, target_index)
        if result is None:
            return self
        index, stop, target_index, post_index = result

        values = self._internal[index:stop]
        internal = self._internal.delete(index, stop)

        if post_index == len(internal):
            return self._make(internal.extend(values))
        elif post_index == 0:
            return self._make(pvector(values) + internal)
        else:
            return self._make(
                internal[:post_index] + pvector(values) + internal[post_index:]
            )

    def _delete(self, item):
        # type: (_LS, Union[slice, int]) -> _LS
        """
        Delete values at index/slice.

        :param item: Index/slice.
        :type item: int or slice

        :return: Transformed.
        :rtype: objetto.states.ListState
        """
        if isinstance(item, slice):
            index, stop = self.resolve_continuous_slice(item)
            return self._make(self._internal.delete(index, stop))
        else:
            index = self.resolve_index(item)
            return self._make(self._internal.delete(index, None))

    def _update(self, index, *values):
        # type: (_LS, int, T) -> _LS
        """
        Update value(s) starting at index.

        :param index: Index.
        :type index: int

        :param values: Value(s).

        :return: Transformed.
        :rtype: objetto.states.ListState

        :raises ValueError: No values provided.
        """
        if not values:
            error = "no values provided"
            raise ValueError(error)
        index = self.resolve_index(index)
        stop = self.resolve_index(index + len(values) - 1) + 1
        pairs = chain.from_iterable(zip(range(index, stop), values))
        new_internal = self._internal.mset(*pairs)  # type: ignore
        return self._make(new_internal)

    def count(self, value):
        # type: (Any) -> int
        """
        Count number of occurrences of a value.

        :param value: Value.

        :return: Number of occurrences.
        :rtype: int
        """
        return self._internal.count(value)

    def index(self, value, start=None, stop=None):
        # type: (Any, Optional[int], Optional[int]) -> int
        """
        Get index of a value.

        :param value: Value.

        :param start: Start index.
        :type start: int or None

        :param stop: Stop index.
        :type stop: int or None

        :return: Index of value.
        :rtype: int

        :raises ValueError: Provided stop but did not provide start.
        """
        if start is None and stop is None:
            args = (value,)  # type: Tuple[Any, ...]
        elif start is not None and stop is None:
            args = (value, start)
        elif start is not None and stop is not None:
            args = (value, start, stop)
        else:
            error = "provided 'stop' argument but did not provide 'start'"
            raise ValueError(error)
        return self._internal.index(*args)

    def resolve_index(self, index, clamp=False):
        # type: (int, bool) -> int
        """
        Resolve index to a positive number.

        :param index: Input index.
        :type index: int

        :param clamp: Whether to clamp between zero and the length.
        :type clamp: bool

        :return: Resolved index.
        :rtype: int

        :raises IndexError: Index out of range.
        """
        return resolve_index(len(self._internal), index, clamp=clamp)

    def resolve_continuous_slice(self, slc):
        # type: (slice) -> Tuple[int, int]
        """
        Resolve continuous slice according to length.

        :param slc: Continuous slice.
        :type slc: slice

        :return: Index and stop.
        :rtype: tuple[int, int]

        :raises IndexError: Slice is noncontinuous.
        """
        return resolve_continuous_slice(len(self._internal), slc)

    def find_with_attributes(self, **attributes):
        # type: (Any) -> T
        """
        Find first value that matches unique attribute values.

        :param attributes: Attributes to match.

        :return: Value that has matching attributes.

        :raises ValueError: No attributes provided or no match found.
        """
        if not attributes:
            error = "no attributes provided"
            raise ValueError(error)
        for value in self._internal:
            for a_name, a_value in iteritems(attributes):
                if not hasattr(value, a_name) or getattr(value, a_name) != a_value:
                    break
            else:
                return value
        error = "could not find a match for {}".format(
            custom_mapping_repr(
                attributes,
                prefix="(",
                template="{key}={value}",
                suffix=")",
                key_repr=str,
            ),
        )
        raise ValueError(error)

    @property
    def _internal(self):
        # type: () -> PVector[T]
        """Internal values."""
        return cast("PVector[T]", super(ListState, self)._internal)



# -*- coding: utf-8 -*-
"""Utilities to retrieve the caller's module name."""

from inspect import getmodule, stack
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Optional

__all__ = ["get_caller_module"]


def get_caller_module(frames=0):
    # type: (int) -> Optional[str]
    """
    Get caller module name if possible.

    :param frames: How many frames in the stack to go back relative to the caller.
    :type frames: int

    :return: Module name or `None`.
    :rtype: str or None
    """
    try:
        frame = stack()[2 + frames]
        module = getmodule(frame[0])
    except IndexError:
        return None
    else:
        if module is None:
            return None
        else:
            return module.__name__



# -*- coding: utf-8 -*-

import pickle

import pytest

from objetto.utils.weak_reference import WeakReference


class Cls(object):
    pass


def test_empty_weak_reference():
    weak = WeakReference()
    assert weak() is None


def test_weak_reference():
    strong = Cls()
    strong_hash = hash(strong)
    weak_a = WeakReference(strong)
    weak_b = WeakReference(strong)
    assert weak_a() is strong
    assert weak_b() is strong
    assert hash(weak_a) == strong_hash
    assert hash(weak_b) == strong_hash
    assert weak_a == weak_b
    del strong
    assert weak_a() is None
    assert weak_b() is None
    assert hash(weak_a) == strong_hash
    assert hash(weak_b) == strong_hash
    assert weak_a == weak_b


def test_pickling():
    strong = Cls()
    weak = WeakReference(strong)
    unpickled_strong, unpickled_weak = pickle.loads(pickle.dumps((strong, weak)))
    assert isinstance(unpickled_strong, Cls)
    assert isinstance(unpickled_weak, WeakReference)
    assert unpickled_weak() is unpickled_strong
    assert type(pickle.loads(pickle.dumps(weak()))) is Cls


if __name__ == "__main__":
    pytest.main()
